---- MODULE MemoirHLSpecification ----

(***************************************************************************)
(* This module defines the high-level behavior of `^\name^'.  There are    *)
(* three actions:                                                          *)
(*     HLMakeInputAvailable                                                *)
(*     HLAdvanceService                                                    *)
(*     HLDie                                                               *)
(***************************************************************************)

EXTENDS MemoirCommon

VARIABLE HLAlive \*`^\label{VARIABLE-HLAlive}^'
VARIABLE HLAvailableInputs \*`^\label{VARIABLE-HLAvailableInputs}^'
VARIABLE HLObservedOutputs \*`^\label{VARIABLE-HLObservedOutputs}^'
VARIABLE HLPublicState \*`^\label{VARIABLE-HLPublicState}^'
VARIABLE HLPrivateState \*`^\label{VARIABLE-HLPrivateState}^'

HLTypeInvariant == \*`^\label{DEF-HLTypeInvariant}^'
    /\  HLAlive \in BOOLEAN
    /\  HLAvailableInputs \subseteq InputType
    /\  HLObservedOutputs \subseteq OutputType
    /\  HLPublicState \in PublicStateType
    /\  HLPrivateState \in PrivateStateType

HLVars == <<HLAlive, HLAvailableInputs, HLObservedOutputs, HLPublicState, HLPrivateState>> \*`^\label{DEF-HLVars}^'

----

(***************************************************************************)
(* The HLAdvanceService action is not allowed to take just any input from  *)
(* InputType.  It may only take an input from the set HLAvailableInputs.   *)
(* This models the fact that some inputs might not be known to the user    *)
(* that invokes the service.  For example, the service might be used to    *)
(* redeem cryptographically signed tokens, and the user does not initially *)
(* know the complete set of valid tokens.  The user might, for example,    *)
(* have to pay money to retrieve a token from a server, and when the user  *)
(* does so, this corresponds to the action HLMakeInputAvailable, which     *)
(* puts the input that includes this token into the set of                 *)
(* HLAvailableInputs.                                                      *)
(***************************************************************************)

HLMakeInputAvailable == \*`^\label{DEF-HLMakeInputAvailable}^'
    \E input \in InputType :
        /\  input \notin HLAvailableInputs
        /\  HLAvailableInputs' = HLAvailableInputs \union {input}
        /\  UNCHANGED HLObservedOutputs
        /\  UNCHANGED HLAlive
        /\  UNCHANGED HLPublicState
        /\  UNCHANGED HLPrivateState

(***************************************************************************)
(* The high-level behavior is a service.  There is one main action, which  *)
(* is HLAdvanceService.  This action takes some input, invokes the         *)
(* developer-supplied service with this input and the current public and   *)
(* private state, updates the public and private state accordingly, and    *)
(* adds the output to the set of observed outputs.                         *)
(***************************************************************************)

HLAdvanceService == \*`^\label{DEF-HLAdvanceService}^'
    \E input \in HLAvailableInputs :
        LET
            hlSResult == Service(HLPublicState, HLPrivateState, input)
        IN
            /\  HLAlive = TRUE
            /\  HLPublicState' = hlSResult.newPublicState
            /\  HLPrivateState' = hlSResult.newPrivateState
            /\  HLObservedOutputs' = HLObservedOutputs \union {hlSResult.output}
            /\  UNCHANGED HLAvailableInputs
            /\  UNCHANGED HLAlive

(***************************************************************************)
(* The high-level spec includes the HLDie action, which kills the system.  *)
(* This is necessary because, in the low-level specs, it is possible for   *)
(* an adversary to perform an action that causes the system to no longer   *)
(* function.  Therefore, we must admit a corresponding action in the       *)
(* high-level spec.  Importantly, the HLDie action does not change the set *)
(* of observed outputs, so it cannot be used to trick the system into      *)
(* providing an output it would not otherwise be willing to provide.       *)
(***************************************************************************)

HLDie == \*`^\label{DEF-HLDie}^'
    /\  HLAlive' = FALSE
    /\  UNCHANGED HLAvailableInputs
    /\  UNCHANGED HLObservedOutputs
    /\  HLPublicState' = DeadPublicState
    /\  HLPrivateState' = DeadPrivateState

----

HLInit == \*`^\label{DEF-HLInit}^'
    /\  HLAlive = TRUE
    /\  HLAvailableInputs = InitialAvailableInputs
    /\  HLObservedOutputs = {}
    /\  HLPublicState = InitialPublicState
    /\  HLPrivateState = InitialPrivateState

HLNext == \*`^\label{DEF-HLNext}^'
    \/  HLMakeInputAvailable
    \/  HLAdvanceService
    \/  HLDie

HLSpec == HLInit /\ [][HLNext]_HLVars \*`^\label{DEF-HLSpec}^'

====
