---- MODULE MemoirLL1Implementation ----

(***************************************************************************)
(* This module proves that the `^\basicname^' spec implements the          *)
(* high-level spec, under the defined refinement.  It begins with two      *)
(* supporting lemmas:                                                      *)
(*     NonAdvancementLemma                                                 *)
(*     LL1NVRAMHistorySummaryUncorruptedEqualsHLAliveLemma                 *)
(* Then, it states and proves the LL1Implementation theorem.               *)
(***************************************************************************)

EXTENDS MemoirLL1InclCardUniqInvariance

(***************************************************************************)
(* The NonAdvancementLemma proves that, if there is no change to the NVRAM *)
(* or to the authentication status of any history state binding, then the  *)
(* high-level public and private state defined by the refinement both      *)
(* stutter.                                                                *)
(***************************************************************************)

THEOREM NonAdvancementLemma == \*`^\label{THEOREM-NonAdvancementLemma}^'
      ( /\  LL1Refinement
        /\  LL1Refinement'
        /\  LL1TypeInvariant
        /\  LL1TypeInvariant'
        /\  UniquenessInvariant
        /\  UNCHANGED LL1NVRAM
        /\  \A historyStateBinding \in HashType :
                UNCHANGED LL1HistoryStateBindingAuthenticated(historyStateBinding) )
    =>
        UNCHANGED <<HLPublicState, HLPrivateState>>
  (*************************************************************************)
  (* We assume the antecedent.                                             *)
  (*************************************************************************)
  <1>1. HAVE /\  LL1Refinement
             /\  LL1Refinement'
             /\  LL1TypeInvariant
             /\  LL1TypeInvariant'
             /\  UniquenessInvariant
             /\  UNCHANGED LL1NVRAM
             /\  \A historyStateBinding \in HashType :
                     UNCHANGED LL1HistoryStateBindingAuthenticated(historyStateBinding)
  <1>2. LL1TypeInvariant /\ LL1TypeInvariant'
    BY <1>1
  (*************************************************************************)
  (* These definitions are copied from the LL1Refinement.                  *)
  (*************************************************************************)
  <1> refPrivateStateEnc == SymmetricEncrypt(LL1NVRAM.symmetricKey, HLPrivateState)
  <1> refStateHash == Hash(HLPublicState, refPrivateStateEnc)
  <1> refHistoryStateBinding == Hash(LL1NVRAM.historySummary, refStateHash)
  (*************************************************************************)
  (* We prove that the definitions satisfy their types in both the         *)
  (* unprimed and primed states, using the LL1RefinementDefsTypeSafeLemma  *)
  (* and the LL1RefinementPrimeDefsTypeSafeLemma.                          *)
  (*************************************************************************)
  <1>3. /\  refPrivateStateEnc \in PrivateStateEncType
        /\  refStateHash \in HashType
        /\  refHistoryStateBinding \in HashType
    BY <1>1, <1>2, LL1RefinementDefsTypeSafeLemma
  <1>4. /\  refPrivateStateEnc' \in PrivateStateEncType
        /\  refStateHash' \in HashType
        /\  refHistoryStateBinding' \in HashType
    BY <1>1, <1>2, LL1RefinementPrimeDefsTypeSafeLemma
  (*************************************************************************)
  (* We then hide the definitions.  We'll pull them in as needed later.    *)
  (*************************************************************************)
  <1> HIDE DEF refPrivateStateEnc, refStateHash, refHistoryStateBinding
  (*************************************************************************)
  (* We consider the two possible states of the                            *)
  (* LL1NVRAMHistorySummaryUncorrupted predicate separately.  In the first  *)
  (* case, that the history state binding in the NVRAM is authenticated in  *)
  (* the unprimed state.                                                   *)
  (*************************************************************************)
  <1>5. CASE LL1NVRAMHistorySummaryUncorrupted
    (***********************************************************************)
    (* We first prove that the history state binding in the NVRAM is        *)
    (* authenticated in the primed state, as well.  This will be needed a  *)
    (* few places below.                                                   *)
    (***********************************************************************)
    <2>1. LL1NVRAMHistorySummaryUncorrupted'
      <3>1. LL1NVRAMHistorySummaryUncorrupted
        BY <1>5
      <3>2. UNCHANGED LL1NVRAMHistorySummaryUncorrupted
        <4>1. /\  LL1TypeInvariant
              /\  UNCHANGED LL1NVRAM
              /\  \A historyStateBinding \in HashType :
                      UNCHANGED LL1HistoryStateBindingAuthenticated(historyStateBinding)
          BY <1>1
        <4>2. QED
         BY <4>1, LL1NVRAMHistorySummaryUncorruptedUnchangedLemma
      <3>3. QED
        BY <3>1, <3>2
    (***********************************************************************)
    (* We show that the refined public state does not change and that the  *)
    (* refined encrypted private state does not change.  We will use the   *)
    (* HashCollisionResistant property.                                    *)
    (***********************************************************************)
    <2>2. /\ UNCHANGED HLPublicState
            /\ UNCHANGED refPrivateStateEnc
      (*********************************************************************)
      (* We begin by proving some types we will need to appeal to the      *)
      (* HashCollisionResistant property.                                  *)
      (*********************************************************************)
      <3>1. HLPublicState \in HashDomain
        <4>1. HLPublicState \in PublicStateType
          <5>1. LL1Refinement
            BY <1>1
          <5>2. QED
            BY <1>5, <5>1 DEF LL1Refinement
        <4>2. QED
          BY <4>1 DEF HashDomain
      <3>2. HLPublicState' \in HashDomain
        <4>1. HLPublicState' \in PublicStateType
          <5>1. LL1Refinement'
            BY <1>1
          <5>2. QED
            BY <2>1, <5>1 DEF LL1Refinement
        <4>2. QED
          BY <4>1 DEF HashDomain
      <3>3. refPrivateStateEnc \in HashDomain
        <4>1. refPrivateStateEnc \in PrivateStateEncType
          BY <1>3
        <4>2. QED
          BY <4>1 DEF HashDomain
      <3>4. refPrivateStateEnc' \in HashDomain
        <4>1. refPrivateStateEnc' \in PrivateStateEncType
          BY <1>4
        <4>2. QED
          BY <4>1 DEF HashDomain
      (*********************************************************************)
      (* We then prove that the refined state hash does not change.  We    *)
      (* will use the UniquenessInvariant.                                 *)
      (*********************************************************************)
      <3>5. UNCHANGED refStateHash
        (*******************************************************************)
        (* We prove some types we will need for the UniquenessInvariant.   *)
        (*******************************************************************)
        <4>1. refStateHash \in HashType
            BY <1>3
        <4>2. refStateHash' \in HashType
            BY <1>4
        (*******************************************************************)
        (* We prove that the unprimed history state binding is              *)
        (* authenticated in the unprimed state.  This follows directly     *)
        (* from the LL1Refinement.                                         *)
        (*******************************************************************)
        <4>3. LL1HistoryStateBindingAuthenticated(refHistoryStateBinding)
          <5>1. LL1Refinement
              BY <1>1
          <5>2. LL1NVRAMHistorySummaryUncorrupted
              BY <1>5
          <5>3. QED
              BY <5>1, <5>2 DEF LL1Refinement, refHistoryStateBinding, refStateHash, refPrivateStateEnc
        (*******************************************************************)
        (* We prove that the primed history state binding is authenticated  *)
        (* in the unprimed state.  This is a little complicated, because   *)
        (* TLA+ does not allow us to prime an operator, only an            *)
        (* expression.  So, we have to expand the definition of            *)
        (* LL1HistoryStateBindingAuthenticated for the four steps of this   *)
        (* proof.                                                          *)
        (*******************************************************************)
        <4>4. LL1HistoryStateBindingAuthenticated(refHistoryStateBinding')
          (*****************************************************************)
          (* Step 1: The primed history state binding is authenticated in   *)
          (* the primed state.  This follows directly from the primed      *)
          (* LL1Refinement.                                                *)
          (*****************************************************************)
          <5>1. \E authenticator \in LL1ObservedAuthenticators' :
                      ValidateMAC(LL1NVRAM.symmetricKey', refHistoryStateBinding', authenticator)
            <6>1. LL1HistoryStateBindingAuthenticated(refHistoryStateBinding)'
              <7>1. LL1Refinement'
                  BY <1>1
              <7>2. LL1NVRAMHistorySummaryUncorrupted'
                BY <2>1
              <7>3. QED
                  BY <7>1, <7>2 DEF LL1Refinement, refHistoryStateBinding, refStateHash, refPrivateStateEnc
            <6>2. QED
                BY <6>1 DEF LL1HistoryStateBindingAuthenticated
          (*****************************************************************)
          (* Step 2: The primed history state binding is of HashType.  We  *)
          (* need this so we can apply the assumption that there is no     *)
          (* change to the authentication status of any history state      *)
          (* binding.                                                      *)
          (*****************************************************************)
          <5>2. refHistoryStateBinding' \in HashType
              BY <1>4
          (*****************************************************************)
          (* Step 3: There is no change to the authentication status of    *)
          (* any history state binding.                                     *)
          (*****************************************************************)
          <5>3. \A historyStateBinding \in HashType :
                        ( \E authenticator \in LL1ObservedAuthenticators' :
                              ValidateMAC(LL1NVRAM.symmetricKey', historyStateBinding, authenticator) )
                      =>
                        ( \E authenticator \in LL1ObservedAuthenticators :
                              ValidateMAC(LL1NVRAM.symmetricKey, historyStateBinding, authenticator) )
            <6>1. \A historyStateBinding \in HashType :
                        UNCHANGED LL1HistoryStateBindingAuthenticated(historyStateBinding)
                BY <1>1
            <6>2. QED
                BY <6>1 DEF LL1HistoryStateBindingAuthenticated
          (*****************************************************************)
          (* Step 4: The primed history state binding is authenticated in   *)
          (* the unprimed state.                                           *)
          (*****************************************************************)
          <5>4. \E authenticator \in LL1ObservedAuthenticators :
                      ValidateMAC(LL1NVRAM.symmetricKey, refHistoryStateBinding', authenticator)
              BY <5>1, <5>2, <5>3
          <5>5. QED
              BY <5>4 DEF LL1HistoryStateBindingAuthenticated
        (*******************************************************************)
        (* Since the authentication status of the history state binding has *)
        (* not changed, the state hash has not changed.                    *)
        (*******************************************************************)
        <4>5. QED
          <5>1. UniquenessInvariant
              BY <1>1
          <5>2. UNCHANGED LL1NVRAM.historySummary
            <6>1. UNCHANGED LL1NVRAM
                BY <1>1
            <6>2. QED
                BY <6>1
          <5>3. QED
              BY <4>1, <4>2, <4>3, <4>4, <5>1, <5>2
              DEF UniquenessInvariant, refHistoryStateBinding
      (*********************************************************************)
      (* Since the refined state hash does not change, the refined public  *)
      (* state and encrypted private state do not change, because the hash *)
      (* is collision-resistant.                                           *)
      (*********************************************************************)
      <3>6. QED
          BY <3>1, <3>2, <3>3, <3>4, <3>5, HashCollisionResistant DEF refStateHash
    (***********************************************************************)
    (* From the previous step, it immediately follows that the refined     *)
    (* public state does not change.                                       *)
    (***********************************************************************)
    <2>3. UNCHANGED HLPublicState 
        BY <2>2
    (***********************************************************************)
    (* It is slightly more involved to show that the refined private state *)
    (* does not change.  We need to show that the symmetric key in the     *)
    (* NVRAM does not change, and we also have to employ the correctness   *)
    (* of the symmetric crypto operations.                                 *)
    (***********************************************************************)
    <2>4. UNCHANGED HLPrivateState
      (*********************************************************************)
      (* In the unprimed state, the refined private state equals the       *)
      (* decryption of the refined encrypted private state, using the      *)
      (* unprimed symmetry key in the NVRAM.  This follows because the     *)
      (* refined encrypted private state is defined in the LL1Refinement   *)
      (* as the encryption of the refined private state, and the symmetric *)
      (* crypto operations are assumed to be correct.                      *)
      (*********************************************************************)
      <3>1. HLPrivateState = SymmetricDecrypt(LL1NVRAM.symmetricKey, refPrivateStateEnc)
        <4>1. LL1NVRAM.symmetricKey \in SymmetricKeyType
            BY <1>2, LL1SubtypeImplicationLemma DEF LL1SubtypeImplication
        <4>2. HLPrivateState \in PrivateStateType
          <5>1. LL1Refinement
            BY <1>1
          <5>2. QED
            BY <1>5, <5>1 DEF LL1Refinement
        <4>3. QED
            BY <4>1, <4>2, SymmetricCryptoCorrect DEF refPrivateStateEnc
      (*********************************************************************)
      (* We then show that in the primed state, the refined private state  *)
      (* equals the decryption of the refined encrypted private state,     *)
      (* using the unprimed symmetry key in the NVRAM.                     *)
      (*********************************************************************)
      <3>2. HLPrivateState' = SymmetricDecrypt(LL1NVRAM.symmetricKey, refPrivateStateEnc)
        (*******************************************************************)
        (* In the primed state, the refined private state equals the       *)
        (* decryption of the refined encrypted private state, using the    *)
        (* primed symmetry key in the NVRAM.  This follows because the     *)
        (* refined encrypted private state is defined in the LL1Refinement *)
        (* as the encryption of the refined private state, and the         *)
        (* symmetric crypto operations are assumed to be correct.          *)
        (*******************************************************************)
        <4>1. HLPrivateState' = SymmetricDecrypt(LL1NVRAM.symmetricKey', refPrivateStateEnc')
          <5>1. LL1NVRAM.symmetricKey' \in SymmetricKeyType
              BY <1>2, LL1SubtypeImplicationLemma DEF LL1SubtypeImplication
          <5>2. HLPrivateState' \in PrivateStateType
            <6>1. LL1Refinement'
              BY <1>1
            <6>2. QED
              BY <2>1, <6>1 DEF LL1Refinement
          <5>3. QED
              BY <5>1, <5>2, SymmetricCryptoCorrect DEF refPrivateStateEnc
        (*******************************************************************)
        (* The symmetry key in the NVRAM is unchanged, and the refined     *)
        (* encrypted private state is unchanged.                           *)
        (*******************************************************************)
        <4>2. UNCHANGED LL1NVRAM.symmetricKey
          <5>1. UNCHANGED LL1NVRAM
              BY <1>1
          <5>2. QED
              BY <5>1
        <4>3. UNCHANGED refPrivateStateEnc
            BY <2>2
        (*******************************************************************)
        (* Since the relevant variables are unchanged, we can conclude     *)
        (* that, in the primed state, the refined private state equals the *)
        (* decryption of the refined encrypted private state, using the    *)
        (* unprimed symmetry key in the NVRAM.                             *)
        (*******************************************************************)
        <4>4. QED
            BY <4>1, <4>2, <4>3
      (*********************************************************************)
      (* Since both the unprimed and primed refined private state are      *)
      (* equal to the same expression, they are equal to each other.       *)
      (*********************************************************************)
      <3>3. QED
          BY <3>1, <3>2
    <2>5. QED
        BY <1>5, <2>3, <2>4
  (*************************************************************************)
  (* In the second case, the history state binding in the NVRAM is not      *)
  (* authenticated in the unprimed state.                                  *)
  (*************************************************************************)
  <1>6. CASE ~LL1NVRAMHistorySummaryUncorrupted
    (***********************************************************************)
    (* By the LL1Refinement, the high-level public and private states are  *)
    (* equal to their dead states.                                         *)
    (***********************************************************************)
    <2>1.  /\  HLPublicState = DeadPublicState
           /\  HLPrivateState = DeadPrivateState
      <3>1. LL1Refinement
        BY <1>1
      <3>2. QED
        BY <1>6, <3>1 DEF LL1Refinement
    (***********************************************************************)
    (* The history state binding in the NVRAM is also not authenticated in  *)
    (* the primed state, so again by the LL1Refinement, the high-level     *)
    (* public and private states are equal to their dead states.           *)
    (***********************************************************************)
    <2>2.  /\  HLPublicState' = DeadPublicState
           /\  HLPrivateState' = DeadPrivateState
      <3>1. ~LL1NVRAMHistorySummaryUncorrupted'
        <4>1. UNCHANGED LL1NVRAMHistorySummaryUncorrupted
          <5>1. /\  LL1TypeInvariant
                  /\  UNCHANGED LL1NVRAM
                  /\  \A historyStateBinding \in HashType :
                          UNCHANGED LL1HistoryStateBindingAuthenticated(historyStateBinding)
              BY <1>1
          <5>2. QED
              BY <5>1, LL1NVRAMHistorySummaryUncorruptedUnchangedLemma
        <4>2. QED
          BY <1>6, <4>1
      <3>2. LL1Refinement'
        BY <1>1
      <3>3. QED
        BY <3>1, <3>2 DEF LL1Refinement
    <2>3. QED
      BY <2>1, <2>2
  <1>7. QED
    BY <1>5, <1>6

----

(***************************************************************************)
(* This simple lemma proves that the two predicates                        *)
(* LL1NVRAMHistorySummaryUncorrupted and HLAlive are equal whenever        *)
(* LL1Refinement is true, in either unprimed or primed state.              *)
(***************************************************************************)

THEOREM LL1NVRAMHistorySummaryUncorruptedEqualsHLAliveLemma == \*`^\label{THEOREM-LL1NVRAMHistorySummaryUncorruptedEqualsHLAliveLemma}^'
    /\ LL1Refinement => LL1NVRAMHistorySummaryUncorrupted = HLAlive
    /\ LL1Refinement' => LL1NVRAMHistorySummaryUncorrupted' = HLAlive'
  <1>1. LL1Refinement => LL1NVRAMHistorySummaryUncorrupted = HLAlive
    <2>1. HAVE LL1Refinement
    <2>2. LL1NVRAMHistorySummaryUncorrupted \in BOOLEAN
      BY DEF LL1NVRAMHistorySummaryUncorrupted
    <2>3. HLAlive \in BOOLEAN
      BY <2>1 DEF LL1Refinement
    <2>4. IF LL1NVRAMHistorySummaryUncorrupted THEN HLAlive = TRUE ELSE HLAlive = FALSE
      BY <2>1 DEF LL1Refinement
    <2>5. QED
      BY <2>2, <2>3, <2>4
  <1>2. LL1Refinement' => LL1NVRAMHistorySummaryUncorrupted' = HLAlive'
    <2>1. HAVE LL1Refinement'
    <2>2. LL1NVRAMHistorySummaryUncorrupted' \in BOOLEAN
      BY DEF LL1NVRAMHistorySummaryUncorrupted
    <2>3. HLAlive' \in BOOLEAN
      BY <2>1 DEF LL1Refinement
    <2>4. IF LL1NVRAMHistorySummaryUncorrupted' THEN HLAlive' = TRUE ELSE HLAlive' = FALSE
      BY <2>1 DEF LL1Refinement
    <2>5. QED
      BY <2>2, <2>3, <2>4
  <1>3. QED
    BY <1>1, <1>2

----

(***************************************************************************)
(* The LL1Implementation theorem is where the rubber meets the road.  This *)
(* is the ultimate proof that the `^\basicname^' spec implements the       *)
(* high-level spec, under the defined refinement.                          *)
(***************************************************************************)

THEOREM LL1Implementation == LL1Spec /\ []LL1Refinement => HLSpec \*`^\label{THEOREM-LL1Implementation}^'
  (*************************************************************************)
  (* This proof will require the LL1TypeInvariant and the                  *)
  (* CorrectnessInvariants.  Fortunately, the LL1TypeSafe theorem has      *)
  (* already proven that the `^\basicname^' spec satisfies its type        *)
  (* invariant, and the CorrectnessInvariance theorem has already proven   *)
  (* that the `^\basicname^' spec satisfies the CorrectnessInvariant.      *)
  (*************************************************************************)
  <1>1. LL1Spec => []LL1TypeInvariant
    BY LL1TypeSafe
  <1>2. LL1Spec => []CorrectnessInvariants
    BY LL1TypeSafe, CorrectnessInvariance
  (*************************************************************************)
  (* The top level of the proof is boilerplate TLA+ for a StepSimulation   *)
  (* proof.  First, we prove that the initial predicate of the             *)
  (* `^\basicname^' spec, conjoined with the LL1Refinement and invariants, *)
  (* implies the initial predicate of the high-level spec.  Second, we     *)
  (* prove that the LL1Next predicate, conjoined with the LL1Refinement    *)
  (* and invariants in both primed and unprimed states, implies the HLNext *)
  (* pridicate.  Third, we use temporal induction to prove that these two  *)
  (* conditions imply that, if the LL1Refinement and the invariants always *)
  (* hold, the LL1Spec implies the HLSpec.                                 *)
  (*************************************************************************)
  <1>3. LL1Init /\ LL1Refinement /\ LL1TypeInvariant /\ CorrectnessInvariants => HLInit
    (***********************************************************************)
    (* We begin the base case by assuming the antecedent.                  *)
    (***********************************************************************)
    <2>1. HAVE LL1Init /\ LL1Refinement /\ LL1TypeInvariant /\ CorrectnessInvariants
    <2>2. LL1TypeInvariant
      BY <2>1
    (***********************************************************************)
    (* We pick a symmetricKey that satisfies the LL1Init predicate.        *)
    (***********************************************************************)
    <2>3. PICK symmetricKey \in SymmetricKeyType : LL1Init!(symmetricKey)!1
      BY <2>1 DEF LL1Init
    (***********************************************************************)
    (* We re-state the definitions from the LL1Refinement.                 *)
    (***********************************************************************)
    <2> refPrivateStateEnc == SymmetricEncrypt(LL1NVRAM.symmetricKey, HLPrivateState)
    <2> refStateHash == Hash(HLPublicState, refPrivateStateEnc)
    <2> refHistoryStateBinding == Hash(LL1NVRAM.historySummary, refStateHash)
    (***********************************************************************)
    (* We prove that the definitions from the LL1Refinement satisfy their  *)
    (* types, using the LL1RefinementDefsTypeSafeLemma.                    *)
    (***********************************************************************)
    <2>4. /\  refPrivateStateEnc \in PrivateStateEncType
          /\  refStateHash \in HashType
          /\  refHistoryStateBinding \in HashType
      BY <2>1, <2>2, LL1RefinementDefsTypeSafeLemma
    (***********************************************************************)
    (* We hide the definitions from the LL1Refinement.                     *)
    (***********************************************************************)
    <2> HIDE DEF refPrivateStateEnc, refStateHash, refHistoryStateBinding
    (***********************************************************************)
    (* We re-state the definitions from LL1Init.                           *)
    (***********************************************************************)
    <2> initialPrivateStateEnc == SymmetricEncrypt(symmetricKey, InitialPrivateState)
    <2> initialStateHash == Hash(InitialPublicState, initialPrivateStateEnc)
    <2> initialHistoryStateBinding == Hash(BaseHashValue, initialStateHash)
    <2> initialAuthenticator == GenerateMAC(symmetricKey, initialHistoryStateBinding)
    <2> initialUntrustedStorage == [
            publicState |-> InitialPublicState,
            privateStateEnc |-> initialPrivateStateEnc,
            historySummary |-> BaseHashValue,
            authenticator |-> initialAuthenticator ]
    <2> initialTrustedStorage == [
            historySummary |-> BaseHashValue,
            symmetricKey |-> symmetricKey ]
    (***********************************************************************)
    (* We prove that the definitions from LL1Init satisfy their types,     *)
    (* using the LL1InitDefsTypeSafeLemma.                                 *)
    (***********************************************************************)
    <2>5. /\  initialPrivateStateEnc \in PrivateStateEncType
          /\  initialStateHash \in HashType
          /\  initialHistoryStateBinding \in HashType
          /\  initialAuthenticator \in MACType
          /\  initialUntrustedStorage \in LL1UntrustedStorageType
          /\  initialTrustedStorage \in LL1TrustedStorageType
      <3>1. symmetricKey \in SymmetricKeyType
        BY <2>3
      <3>2. QED
        BY <2>2, <3>1, LL1InitDefsTypeSafeLemma
    (***********************************************************************)
    (* We hide the definitions from LL1Init.                               *)
    (***********************************************************************)
    <2> HIDE DEF initialPrivateStateEnc, initialStateHash, initialHistoryStateBinding,
            initialAuthenticator, initialUntrustedStorage, initialTrustedStorage
    (***********************************************************************)
    (* A couple of the steps below require knowing that the initial        *)
    (* history state binding is authenticated.                             *)
    (***********************************************************************)
    <2>6. LL1HistoryStateBindingAuthenticated(initialHistoryStateBinding)
      (*********************************************************************)
      (* The initial authenticator was genererated as a MAC of the  *)
      (* initial history state binding by LL1Init, using a symmetric key    *)
      (* that matches the symmetric key in the NVRAM.                      *)
      (*********************************************************************)
      <3>1. initialAuthenticator = GenerateMAC(LL1NVRAM.symmetricKey, initialHistoryStateBinding)
        <4>1. LL1NVRAM.symmetricKey = symmetricKey
          <5>1. LL1NVRAM = [ historySummary |-> BaseHashValue,
                          symmetricKey |-> symmetricKey ]
            BY <2>3
          <5>2. QED
            BY <5>1
        <4>2. QED
          BY <2>3, <4>1
          DEF initialAuthenticator, initialHistoryStateBinding,
              initialStateHash, initialPrivateStateEnc
      (*********************************************************************)
      (* We can thus use the MACComplete property to show that the         *)
      (* generated MAC validates appropriately.  To do this, we first need *)
      (* to prove some types.                                              *)
      (*********************************************************************)
      <3>2. LL1NVRAM.symmetricKey \in SymmetricKeyType
        BY <2>2, LL1SubtypeImplicationLemma DEF LL1SubtypeImplication
      <3>3. initialHistoryStateBinding \in HashType
        BY <2>5
      (*********************************************************************)
      (* Then, we appeal to the MACComplete property in a straightforward  *)
      (* way.                                                              *)
      (*********************************************************************)
      <3>4. ValidateMAC(LL1NVRAM.symmetricKey, initialHistoryStateBinding, initialAuthenticator)
        BY <3>1, <3>2, <3>3, MACComplete
      <3>5. initialAuthenticator \in LL1ObservedAuthenticators
        <4>1. LL1ObservedAuthenticators = {initialAuthenticator}
          BY <2>3
          DEF initialAuthenticator, initialHistoryStateBinding,
              initialStateHash, initialPrivateStateEnc
        <4>2. QED
          BY <4>1
      <3>6. QED
        BY <3>4, <3>5 DEF LL1HistoryStateBindingAuthenticated 
    (***********************************************************************)
    (* For a couple of the steps below, we will need to know that the      *)
    (* history state binding in the NVRAM is true.                          *)
    (***********************************************************************)
    <2>7. LL1NVRAMHistorySummaryUncorrupted
      (*********************************************************************)
      (* The definition of LL1NVRAMHistorySummaryUncorrupted asserts that  *)
      (* there is some state hash bound to the history summary in the      *)
      (* NVRAM by a history state binding that is authenticated.  Our      *)
      (* witness for the state hash is the initial state hash.             *)
      (*********************************************************************)
      <3>1. initialStateHash \in HashType
        BY <2>5
      (*********************************************************************)
      (* The initial state hash is bound to the base hash value by the     *)
      (* initial history state binding, and the base hash value is the      *)
      (* initial value of the history summary in the NVRAM.                 *)
      (*********************************************************************)
      <3>2. LL1NVRAM.historySummary = BaseHashValue
        <4>1. LL1NVRAM = [ historySummary |-> BaseHashValue,
                        symmetricKey |-> symmetricKey ]
          BY <2>3
        <4>2. QED
          BY <4>1
      (*********************************************************************)
      (* The initial history state binding is authenticated.                *)
      (*********************************************************************)
      <3>3. LL1HistoryStateBindingAuthenticated(initialHistoryStateBinding)
        BY <2>6
      <3>4. QED
        BY <3>1, <3>2, <3>3 DEF initialHistoryStateBinding, LL1NVRAMHistorySummaryUncorrupted
    (***********************************************************************)
    (* The bulk of the proof for the initial case is proving that the      *)
    (* public and private state defined by the refinememt match the        *)
    (* initial public and private state.                                   *)
    (***********************************************************************)
    <2>8. /\  HLPublicState = InitialPublicState
          /\  HLPrivateState = InitialPrivateState
      (*********************************************************************)
      (* We show that the refined public state matches the initial public  *)
      (* state that the refined encrypted private state matches the        *)
      (* encrypted initial state.  We will use the HashCollisionResistant  *)
      (* property.                                                         *)
      (*********************************************************************)
      <3>1. /\ HLPublicState = InitialPublicState
            /\ refPrivateStateEnc = initialPrivateStateEnc
        (*******************************************************************)
        (* We begin by proving some types we will need to appeal to the    *)
        (* HashCollisionResistant property.                                *)
        (*******************************************************************)
        <4>1. HLPublicState \in HashDomain
          <5>1. HLPublicState \in PublicStateType
            BY <2>1, <2>7 DEF LL1Refinement
          <5>2. QED
            BY <5>1 DEF HashDomain
        <4>2. InitialPublicState \in HashDomain
          <5>1. InitialPublicState \in PublicStateType
            BY ConstantsTypeSafe
          <5>2. QED
            BY <5>1 DEF HashDomain
        <4>3. refPrivateStateEnc \in HashDomain
          <5>1. refPrivateStateEnc \in PrivateStateEncType
            BY <2>4
          <5>2. QED
            BY <5>1 DEF HashDomain
        <4>4. initialPrivateStateEnc \in HashDomain
          <5>1. initialPrivateStateEnc \in PrivateStateEncType
            BY <2>5
          <5>2. QED
            BY <5>1 DEF HashDomain
        (*******************************************************************)
        (* We then prove that the refined state hash equals the initial    *)
        (* state hash.  We will use the UniquenessInvariant.               *)
        (*******************************************************************)
        <4>5. refStateHash = initialStateHash
          (*****************************************************************)
          (* We prove some types we will need for the UniquenessInvariant. *)
          (*****************************************************************)
          <5>1. refStateHash \in HashType
            BY <2>4
          <5>2. initialStateHash \in HashType
            BY <2>5
          (*****************************************************************)
          (* We prove that the refined history state binding is             *)
          (* authenticated.  This follows directly from the LL1Refinement. *)
          (*****************************************************************)
          <5>3. LL1HistoryStateBindingAuthenticated(refHistoryStateBinding)
            <6>1. LL1Refinement
              BY <2>1
            <6>2. LL1NVRAMHistorySummaryUncorrupted
              BY <2>7
            <6>3. QED
              BY <6>1, <6>2 DEF LL1Refinement, refHistoryStateBinding, refStateHash, refPrivateStateEnc
          (*****************************************************************)
          (* We prove that the initial history state binding is             *)
          (* authenticated.  We will use the MACComplete property.         *)
          (*****************************************************************)
          <5>4. LL1HistoryStateBindingAuthenticated(initialHistoryStateBinding)
            BY <2>6
          (*****************************************************************)
          (* Since both history state bindings are authenticated, it        *)
          (* follows that the two state hashes are equal.                  *)
          (*****************************************************************)
          <5>5. QED
            <6>1. UniquenessInvariant
              BY <2>1 DEF CorrectnessInvariants
            (***************************************************************)
            (* For the UniquenessInvariant to apply, we have to show that  *)
            (* the history summary in the initial history state binding is   *)
            (* equal to the history summary in the NVRAM.                   *)
            (***************************************************************)
            <6>2. LL1NVRAM.historySummary = BaseHashValue
              <8>1. LL1NVRAM = [ historySummary |-> BaseHashValue,
                              symmetricKey |-> symmetricKey ]
                BY <2>3
              <8>2. QED
                BY <8>1
            <6>3. QED
              BY <5>1, <5>2, <5>3, <5>4, <6>1, <6>2
              DEF UniquenessInvariant, refHistoryStateBinding, initialHistoryStateBinding
        <4>6. QED
          BY <4>1, <4>2, <4>3, <4>4, <4>5, HashCollisionResistant DEF refStateHash, initialStateHash
      (*********************************************************************)
      (* From the previous step, it immediately follows that the refined   *)
      (* public state matches the initial public state.                    *)
      (*********************************************************************)
      <3>2. HLPublicState = InitialPublicState
        BY <3>1
      (*********************************************************************)
      (* It is slightly more involved to show that the refined private     *)
      (* state matches the initial private state.  We need to show that    *)
      (* the symmetric key that satisfies the LL1Init predicate is the     *)
      (* same symmetric key that is in the NVRAM, and we also have to      *)
      (* employ the correctness of the symmetric crypto operations.        *)
      (*********************************************************************)
      <3>3. HLPrivateState = InitialPrivateState
        (*******************************************************************)
        (* We need to prove some types.                                    *)
        (*******************************************************************)
        <4>1. LL1NVRAM.symmetricKey \in SymmetricKeyType
          BY <2>2, LL1SubtypeImplicationLemma DEF LL1SubtypeImplication
        <4>2. refPrivateStateEnc = initialPrivateStateEnc
          BY <3>1
        (*******************************************************************)
        (* The refined private state equals the decryption of the refined  *)
        (* encrypted private state, using the symmetry key brought into    *)
        (* existence in LL1Init.                                           *)
        (*******************************************************************)
        <4>3. HLPrivateState = SymmetricDecrypt(symmetricKey, refPrivateStateEnc)
          (*****************************************************************)
          (* The refined private state equals the decryption of the        *)
          (* refined encrypted private state, using the symmetry key in    *)
          (* the NVRAM.  This follows because the refined encrypted        *)
          (* private state is defined in the LL1Refinement as the          *)
          (* encryption of the refined private state, and the symmetric    *)
          (* crypto operations are assumed to be correct.                  *)
          (*****************************************************************)
          <5>1. HLPrivateState = SymmetricDecrypt(LL1NVRAM.symmetricKey, refPrivateStateEnc)
            <6>1. HLPrivateState \in PrivateStateType
              BY <2>1, <2>7 DEF LL1Refinement
            <6>2. QED
              BY <4>1, <6>1, SymmetricCryptoCorrect DEF refPrivateStateEnc
          (*****************************************************************)
          (* The symmetric key in the NVRAM matches the symmetric key      *)
          (* brought into existence in LL1Init.                            *)
          (*****************************************************************)
          <5>2. LL1NVRAM.symmetricKey = symmetricKey
            <6>1. LL1NVRAM = [ historySummary |-> BaseHashValue,
                            symmetricKey |-> symmetricKey ]
              BY <2>3
            <6>2. QED
              BY <6>1
          <5>3. QED
            BY <5>1, <5>2
        (*******************************************************************)
        (* The initial private state equals the decryption of the initial  *)
        (* encrypted private state, using the symmetry key brought into    *)
        (* existence in LL1Init.  This follows because the refined         *)
        (* encrypted private state is defined in LL1Init as the encryption *)
        (* of the initial private state, and the symmetric crypto          *)
        (* operations are assumed to be correct.                           *)
        (*******************************************************************)
        <4>4. InitialPrivateState = SymmetricDecrypt(symmetricKey, initialPrivateStateEnc)
          <5>1. InitialPrivateState \in PrivateStateType
            BY ConstantsTypeSafe
          <5>2. QED
            BY <4>1, <5>1, SymmetricCryptoCorrect DEF initialPrivateStateEnc
        <4>5. QED
          BY <4>2, <4>3, <4>4
      (*********************************************************************)
      (* The truth of the conjuncts implies the truth of the conjunction.  *)
      (*********************************************************************)
      <3>4. QED
        BY <3>2, <3>3
    (***********************************************************************)
    (* The QED step simply asserts each conjunct in the HLinit predicate.  *)
    (***********************************************************************)
    <2>9. QED
      <3>1. HLAlive = TRUE
        <4>1. LL1NVRAMHistorySummaryUncorrupted
          BY <2>7
        <4>2. QED
          BY <2>1, <4>1 DEF LL1Refinement
      <3>2. HLAvailableInputs = InitialAvailableInputs
        <4>1. LL1AvailableInputs = InitialAvailableInputs
          BY <2>3
        <4>2. HLAvailableInputs = LL1AvailableInputs
          BY <2>1 DEF LL1Refinement
        <4>3. QED
          BY <4>1, <4>2
      <3>3. HLObservedOutputs = {}
        <4>1. LL1ObservedOutputs = {}
          BY <2>3
        <4>2. HLObservedOutputs = LL1ObservedOutputs
          BY <2>1 DEF LL1Refinement
        <4>3. QED
          BY <4>1, <4>2
      <3>4. HLPublicState = InitialPublicState
        BY <2>8
      <3>5. HLPrivateState = InitialPrivateState
        BY <2>8
      <3>6. QED
        BY <3>1, <3>2, <3>3, <3>4, <3>5 DEF HLInit
  (*************************************************************************)
  (* For the induction step, we will need the refinement, the type         *)
  (* invariant, and the correctness invariants to be true in both the      *)
  (* unprimed and primed states.                                           *)
  (*************************************************************************)
  <1>4. ( /\  [LL1Next]_LL1Vars
          /\  LL1Refinement
          /\  LL1Refinement'
          /\  LL1TypeInvariant
          /\  LL1TypeInvariant'
          /\  CorrectnessInvariants
          /\  CorrectnessInvariants' )
      =>
          [HLNext]_HLVars
    (***********************************************************************)
    (* We assume the antecedents.                                          *)
    (***********************************************************************)
    <2>1. HAVE /\  [LL1Next]_LL1Vars
               /\  LL1Refinement
               /\  LL1Refinement'
               /\  LL1TypeInvariant
               /\  LL1TypeInvariant'
               /\  CorrectnessInvariants
               /\  CorrectnessInvariants'
    (***********************************************************************)
    (* We then prove that each step in the `^\basicname^' spec refines to  *)
    (* a step in the high-level spec.  First, a `^\basicname^' stuttering  *)
    (* step refines to a high-level stuttering step.                       *)
    (***********************************************************************)
    <2>2. UNCHANGED LL1Vars => UNCHANGED HLVars
      <3>1. HAVE UNCHANGED LL1Vars
      (*********************************************************************)
      (* The HLAlive predicate is unchanged because the                    *)
      (* LL1NVRAMHistorySummaryUncorrupted predicate is unchanged.          *)
      (*********************************************************************)
      <3>2. UNCHANGED HLAlive
        (*******************************************************************)
        (* The LL1NVRAMHistorySummaryUncorruptedEqualsHLAliveLemma tells us *)
        (* that LL1NVRAMHistorySummaryUncorrupted and HLAlive are equal.    *)
        (*******************************************************************)
        <4>1. LL1NVRAMHistorySummaryUncorrupted = HLAlive
          <5>1. LL1Refinement
            BY <2>1
          <5>2. QED
            BY <5>1, LL1NVRAMHistorySummaryUncorruptedEqualsHLAliveLemma
        <4>2. LL1NVRAMHistorySummaryUncorrupted' = HLAlive'
          <5>1. LL1Refinement'
            BY <2>1
          <5>2. QED
            BY <5>1, LL1NVRAMHistorySummaryUncorruptedEqualsHLAliveLemma
        (*******************************************************************)
        (* Then, the LL1NVRAMHistorySummaryUncorruptedUnchangedLemma tells  *)
        (* us that the LL1NVRAMHistorySummaryUncorrupted predicate is       *)
        (* unchanged.                                                      *)
        (*******************************************************************)
        <4>3. UNCHANGED LL1NVRAMHistorySummaryUncorrupted
          <5>1. LL1TypeInvariant
            BY <2>1
          <5>2. UNCHANGED LL1NVRAM
            BY <3>1 DEF LL1Vars
          <5>3. \A historyStateBinding \in HashType :
                    UNCHANGED LL1HistoryStateBindingAuthenticated(historyStateBinding)
            <6>1. UNCHANGED <<LL1NVRAM, LL1ObservedAuthenticators>>
              BY <3>1 DEF LL1Vars
            <6>2. QED
              BY <6>1, UnchangedAuthenticatedHistoryStateBindingsLemma
          <5>4. QED
            BY <5>1, <5>2, <5>3, LL1NVRAMHistorySummaryUncorruptedUnchangedLemma
        <4>4. QED
          BY <4>1, <4>2, <4>3
      (*********************************************************************)
      (* The mapping from LL1AvailableInputs to HLAvailableInputs is       *)
      (* direct.                                                           *)
      (*********************************************************************)
      <3>3. UNCHANGED HLAvailableInputs
        <4>1. UNCHANGED LL1AvailableInputs
          BY <3>1 DEF LL1Vars
        <4>2. /\ HLAvailableInputs = LL1AvailableInputs
              /\ HLAvailableInputs' = LL1AvailableInputs'
          BY <2>1 DEF LL1Refinement
        <4>3. QED
          BY <4>1, <4>2
      (*********************************************************************)
      (* The mapping from LL1ObservedOutputs to HLObservedOutputs is       *)
      (* direct.                                                           *)
      (*********************************************************************)
      <3>4. UNCHANGED HLObservedOutputs
        <4>1. UNCHANGED LL1ObservedOutputs
          BY <3>1 DEF LL1Vars
        <4>2. /\ HLObservedOutputs = LL1ObservedOutputs
              /\ HLObservedOutputs' = LL1ObservedOutputs'
          BY <2>1 DEF LL1Refinement
        <4>3. QED
          BY <4>1, <4>2
      (*********************************************************************)
      (* We prove the stuttering of the high-level public and private      *)
      (* state by using the NonAdvancementLemma.                           *)
      (*********************************************************************)
      <3>5. UNCHANGED <<HLPublicState, HLPrivateState>>
        (*******************************************************************)
        (* Many of the antecedents for the NonAdvancementLemma come        *)
        (* directly from antecedents in the induction.                     *)
        (*******************************************************************)
        <4>1. /\  LL1Refinement
              /\  LL1Refinement'
              /\  LL1TypeInvariant
              /\  LL1TypeInvariant'
              /\  UniquenessInvariant
            BY <2>1 DEF CorrectnessInvariants
        (*******************************************************************)
        (* The LL1NVRAM is unchanged because the `^\basicname^' variables  *)
        (* are unchanged.                                                  *)
        (*******************************************************************)
        <4>2. UNCHANGED LL1NVRAM
            BY <3>1 DEF LL1Vars
        (*******************************************************************)
        (* The UnchangedAuthenticatedHistoryStateBindingsLemma tells us     *)
        (* that there is no change to the set of history state bindings     *)
        (* that have authenticators in the set                            *)
        (* LL1ObservedAuthenticators.                                *)
        (*******************************************************************)
        <4>3. \A historyStateBinding \in HashType :
                    UNCHANGED LL1HistoryStateBindingAuthenticated(historyStateBinding)
          <5>1. UNCHANGED <<LL1NVRAM, LL1ObservedAuthenticators>>
              BY <3>1 DEF LL1Vars
          <5>2. QED
              BY <5>1, UnchangedAuthenticatedHistoryStateBindingsLemma
        (*******************************************************************)
        (* We have all of the antecedents for the NonAdvancementLemma, so  *)
        (* we can apply it directly.                                       *)
        (*******************************************************************)
        <4>4. QED
            BY <4>1, <4>2, <4>3, NonAdvancementLemma
      <3>6. QED
        BY <3>2, <3>3, <3>4, <3>5 DEF HLVars
    (***********************************************************************)
    (* A `^\basicname^' LL1MakeInputAvailable action refines to a          *)
    (* high-level HLMakeInputAvailable action.                             *)
    (***********************************************************************)
    <2>3. LL1MakeInputAvailable => HLMakeInputAvailable
      <3>1. HAVE LL1MakeInputAvailable
      <3>2. PICK input \in InputType : LL1MakeInputAvailable!(input)
        BY <3>1 DEF LL1MakeInputAvailable
      (*********************************************************************)
      (* The mapping from LL1AvailableInputs to HLAvailableInputs is       *)
      (* direct.                                                           *)
      (*********************************************************************)
      <3>3. input \notin HLAvailableInputs
        <4>1. input \notin LL1AvailableInputs
          BY <3>2
        <4>2. /\ HLAvailableInputs = LL1AvailableInputs
              /\ HLAvailableInputs' = LL1AvailableInputs'
          BY <2>1 DEF LL1Refinement
        <4>3. QED
          BY <4>1, <4>2
      <3>4. HLAvailableInputs' = HLAvailableInputs \union {input}
        <4>1. LL1AvailableInputs' = LL1AvailableInputs \union {input}
          BY <3>2
        <4>2. /\ HLAvailableInputs = LL1AvailableInputs
              /\ HLAvailableInputs' = LL1AvailableInputs'
          BY <2>1 DEF LL1Refinement
        <4>3. QED
          BY <4>1, <4>2
      (*********************************************************************)
      (* The mapping from LL1ObservedOutputs to HLObservedOutputs is       *)
      (* direct.                                                           *)
      (*********************************************************************)
      <3>5. UNCHANGED HLObservedOutputs
        <4>1. UNCHANGED LL1ObservedOutputs
          BY <3>2
        <4>2. /\ HLObservedOutputs = LL1ObservedOutputs
              /\ HLObservedOutputs' = LL1ObservedOutputs'
          BY <2>1 DEF LL1Refinement
        <4>3. QED
          BY <4>1, <4>2
      (*********************************************************************)
      (* The HLAlive predicate is unchanged because the                    *)
      (* LL1NVRAMHistorySummaryUncorrupted predicate is unchanged.          *)
      (*********************************************************************)
      <3>6. UNCHANGED HLAlive
        (*******************************************************************)
        (* The LL1NVRAMHistorySummaryUncorruptedEqualsHLAliveLemma tells us *)
        (* that LL1NVRAMHistorySummaryUncorrupted and HLAlive are equal.    *)
        (*******************************************************************)
        <4>1. LL1NVRAMHistorySummaryUncorrupted = HLAlive
          <5>1. LL1Refinement
            BY <2>1
          <5>2. QED
            BY <5>1, LL1NVRAMHistorySummaryUncorruptedEqualsHLAliveLemma
        <4>2. LL1NVRAMHistorySummaryUncorrupted' = HLAlive'
          <5>1. LL1Refinement'
            BY <2>1
          <5>2. QED
            BY <5>1, LL1NVRAMHistorySummaryUncorruptedEqualsHLAliveLemma
        (*******************************************************************)
        (* Then, the LL1NVRAMHistorySummaryUncorruptedUnchangedLemma tells  *)
        (* us that the LL1NVRAMHistorySummaryUncorrupted predicate is       *)
        (* unchanged.                                                      *)
        (*******************************************************************)
        <4>3. UNCHANGED LL1NVRAMHistorySummaryUncorrupted
          <5>1. LL1TypeInvariant
            BY <2>1
          <5>2. UNCHANGED LL1NVRAM
            BY <3>2
          <5>3. \A historyStateBinding \in HashType :
                    UNCHANGED LL1HistoryStateBindingAuthenticated(historyStateBinding)
            <6>1. UNCHANGED <<LL1NVRAM, LL1ObservedAuthenticators>>
              BY <3>2
            <6>2. QED
              BY <6>1, UnchangedAuthenticatedHistoryStateBindingsLemma
          <5>4. QED
            BY <5>1, <5>2, <5>3, LL1NVRAMHistorySummaryUncorruptedUnchangedLemma
        <4>4. QED
          BY <4>1, <4>2, <4>3 
      (*********************************************************************)
      (* We prove the stuttering of the high-level public and private      *)
      (* state by using the NonAdvancementLemma.                           *)
      (*********************************************************************)
      <3>7. UNCHANGED <<HLPublicState, HLPrivateState>>
        (*******************************************************************)
        (* Many of the antecedents for the NonAdvancementLemma come        *)
        (* directly from antecedents in the induction.                     *)
        (*******************************************************************)
        <4>1. /\  LL1Refinement
              /\  LL1Refinement'
              /\  LL1TypeInvariant
              /\  LL1TypeInvariant'
              /\  UniquenessInvariant
          BY <2>1 DEF CorrectnessInvariants
        (*******************************************************************)
        (* The NVRAM is unchanged by definition of the                     *)
        (* LL1MakeInputAvailable action.                                   *)
        (*******************************************************************)
        <4>2. UNCHANGED LL1NVRAM
          BY <3>2
        (*******************************************************************)
        (* The UnchangedAuthenticatedHistoryStateBindingsLemma tells us     *)
        (* that there is no change to the set of history state bindings     *)
        (* that have authenticators in the set                            *)
        (* LL1ObservedAuthenticators.                                *)
        (*******************************************************************)
        <4>3. \A historyStateBinding \in HashType :
                  UNCHANGED LL1HistoryStateBindingAuthenticated(historyStateBinding)
          <5>1. UNCHANGED <<LL1NVRAM, LL1ObservedAuthenticators>>
            BY <3>2
          <5>2. QED
            BY <5>1, UnchangedAuthenticatedHistoryStateBindingsLemma
        (*******************************************************************)
        (* We have all of the antecedents for the NonAdvancementLemma, so  *)
        (* we can apply it directly.                                       *)
        (*******************************************************************)
        <4>4. QED
          BY <4>1, <4>2, <4>3, NonAdvancementLemma
      <3>8. QED
        BY <3>2, <3>3, <3>4, <3>5, <3>6, <3>7 DEF HLMakeInputAvailable
    (***********************************************************************)
    (* A `^\basicname^' LL1PerformOperation action refines to a high-level *)
    (* HLAdvanceService action.                                            *)
    (***********************************************************************)
    <2>4. LL1PerformOperation => HLAdvanceService
      (*********************************************************************)
      (* We assume the antecedent.                                         *)
      (*********************************************************************)
      <3>1. HAVE LL1PerformOperation
      (*********************************************************************)
      (* We pick an input that satisfies the LL1PerformOperation           *)
      (* predicate.                                                        *)
      (*********************************************************************)
      <3>2. PICK input \in LL1AvailableInputs : LL1PerformOperation!(input)!1
        BY <3>1 DEF LL1PerformOperation
      (*********************************************************************)
      (* We re-state the definitions from the LL1Refinement                *)
      (*********************************************************************)
      <3> refPrivateStateEnc == SymmetricEncrypt(LL1NVRAM.symmetricKey, HLPrivateState)
      <3> refStateHash == Hash(HLPublicState, refPrivateStateEnc)
      <3> refHistoryStateBinding == Hash(LL1NVRAM.historySummary, refStateHash)
      (*********************************************************************)
      (* We prove that the definitions from the LL1Refinement satisfy      *)
      (* their types, using the LL1RefinementDefsTypeSafeLemma.            *)
      (*********************************************************************)
      <3>3. /\  refPrivateStateEnc \in PrivateStateEncType
            /\  refStateHash \in HashType
            /\  refHistoryStateBinding \in HashType
        <4>1. LL1Refinement
          BY <2>1
        <4>2. LL1TypeInvariant
          BY <2>1
        <4>3. QED
          BY <4>1, <4>2, LL1RefinementDefsTypeSafeLemma
      (*********************************************************************)
      (* We also prove that the primed definitions from the LL1Refinement  *)
      (* satisfy their types, using the                                    *)
      (* LL1RefinementPrimeDefsTypeSafeLemma.                              *)
      (*********************************************************************)
      <3>4. /\  refPrivateStateEnc' \in PrivateStateEncType
            /\  refStateHash' \in HashType
            /\  refHistoryStateBinding' \in HashType
        <4>1. LL1Refinement'
          BY <2>1
        <4>2. LL1TypeInvariant'
          BY <2>1
        <4>3. QED
          BY <4>1, <4>2, LL1RefinementPrimeDefsTypeSafeLemma
      (*********************************************************************)
      (* We hide the definitions from the LL1Refinement.                   *)
      (*********************************************************************)
      <3> HIDE DEF refPrivateStateEnc, refStateHash, refHistoryStateBinding
      (*********************************************************************)
      (* We re-state the definitions from LL1PerformOperation.             *)
      (*********************************************************************)
      <3> stateHash == Hash(LL1RAM.publicState, LL1RAM.privateStateEnc)
      <3> historyStateBinding == Hash(LL1RAM.historySummary, stateHash)
      <3> privateState == SymmetricDecrypt(LL1NVRAM.symmetricKey, LL1RAM.privateStateEnc)
      <3> sResult == Service(LL1RAM.publicState, privateState, input)
      <3> newPrivateStateEnc ==
              SymmetricEncrypt(LL1NVRAM.symmetricKey, sResult.newPrivateState)
      <3> newHistorySummary == Hash(LL1NVRAM.historySummary, input)
      <3> newStateHash == Hash(sResult.newPublicState, newPrivateStateEnc)
      <3> newHistoryStateBinding == Hash(newHistorySummary, newStateHash)
      <3> newAuthenticator == GenerateMAC(LL1NVRAM.symmetricKey, newHistoryStateBinding)
      (*********************************************************************)
      (* We prove that the definitions from LL1PerformOperation satisfy    *)
      (* their types, using the LL1PerformOperationDefsTypeSafeLemma.      *)
      (*********************************************************************)
      <3>5. /\  stateHash \in HashType
            /\  historyStateBinding \in HashType
            /\  privateState \in PrivateStateType
            /\  sResult \in ServiceResultType
            /\  sResult.newPublicState \in PublicStateType
            /\  sResult.newPrivateState \in PrivateStateType
            /\  sResult.output \in OutputType
            /\  newPrivateStateEnc \in PrivateStateEncType
            /\  newHistorySummary \in HashType
            /\  newStateHash \in HashType
            /\  newHistoryStateBinding \in HashType
            /\  newAuthenticator \in MACType
        <4>1. input \in LL1AvailableInputs
          BY <3>2
        <4>2. LL1TypeInvariant
          BY <2>1
        <4>3. QED
          BY <4>1, <4>2, LL1PerformOperationDefsTypeSafeLemma
      (*********************************************************************)
      (* We hide the definitions from LL1PerformOperation.                 *)
      (*********************************************************************)
      <3> HIDE DEF stateHash, historyStateBinding,  privateState, sResult, newPrivateStateEnc,
              newHistorySummary, newStateHash, newHistoryStateBinding, newAuthenticator
      (*********************************************************************)
      (* We re-state the definition from HLAdvanceService.            *)
      (*********************************************************************)
      <3> hlSResult == Service(HLPublicState, HLPrivateState, input)
      (*********************************************************************)
      (* We hide the definition from HLAdvanceService.                *)
      (*********************************************************************)
      <3> HIDE DEF hlSResult
      (*********************************************************************)
      (* We prove that the history summary in the NVRAM is uncorrupted in   *)
      (* the unprimed state.  This follows from the enablement conditions  *)
      (* of the LL1PerformOperation action.  We will expand the            *)
      (* definitions of LL1NVRAMHistorySummaryUncorrupted and               *)
      (* LL1HistoryStateBindingAuthenticated, and prove that the required   *)
      (* conditions are all satisfied.                                     *)
      (*********************************************************************)
      <3>6. LL1NVRAMHistorySummaryUncorrupted
        (*******************************************************************)
        (* The authenticator in the RAM authenticates the history state    *)
        (* binding, since this is an enablement condition of the           *)
        (* LL1PerformOperation action.                                     *)
        (*******************************************************************)
        <4>1. ValidateMAC(LL1NVRAM.symmetricKey, historyStateBinding, LL1RAM.authenticator)
          BY <3>2 DEF historyStateBinding, stateHash
        (*******************************************************************)
        (* The state hash defined in the LL1PerformOperation is in         *)
        (* HashType.                                                       *)
        (*******************************************************************)
        <4>2. stateHash \in HashType
          BY <3>5
        (*******************************************************************)
        (* From the UnforgeabilityInvariant, the authenticator in   *)
        (* the RAM is in the set of observed authenticators.        *)
        (*******************************************************************)
        <4>3. LL1RAM.authenticator \in LL1ObservedAuthenticators
          <5>1. UnforgeabilityInvariant
            BY <2>1 DEF CorrectnessInvariants
          <5>2. historyStateBinding \in HashType
            BY <3>5
          <5>3. QED
            BY <4>1, <5>1, <5>2 DEF UnforgeabilityInvariant
        (*******************************************************************)
        (* The above three conditions are sufficient to satisfy the        *)
        (* LL1NVRAMHistorySummaryUncorrupted predicate in the unprimed      *)
        (* state, given that the history summary in the RAM equals the      *)
        (* history summary in the NVRAM.                                    *)
        (*******************************************************************)
        <4>4. LL1NVRAM.historySummary = LL1RAM.historySummary
          BY <3>2
        <4>5. QED
          BY <4>1, <4>2, <4>3, <4>4
          DEF LL1NVRAMHistorySummaryUncorrupted, LL1HistoryStateBindingAuthenticated,
              historyStateBinding
      (*********************************************************************)
      (* We prove that the history summary in the NVRAM is uncorrupted in   *)
      (* the primed state.  This follows from the enablement conditions of *)
      (* the LL1PerformOperation action.  We will expand the definitions   *)
      (* of LL1NVRAMHistorySummaryUncorrupted and                           *)
      (* LL1HistoryStateBindingAuthenticated, and prove that the required   *)
      (* conditions are all satisfied.                                     *)
      (*********************************************************************)
      <3>7. LL1NVRAMHistorySummaryUncorrupted'
        (*******************************************************************)
        (* The new authenticator defined by the LL1PerformOperation *)
        (* action authenticates the new history state binding defined by    *)
        (* this action.  We will use the MACComplete property.             *)
        (*******************************************************************)
        <4>1. ValidateMAC(LL1NVRAM.symmetricKey', newHistoryStateBinding, newAuthenticator)
          (*****************************************************************)
          (* The new authenticator was genererated as a MAC of the  *)
          (* new history state binding by LL1PerformOperation, using the    *)
          (* unchanged symmetric key in the NVRAM.                         *)
          (*****************************************************************)
          <5>1. newAuthenticator = GenerateMAC(LL1NVRAM.symmetricKey', newHistoryStateBinding)
            <6>1. UNCHANGED LL1NVRAM.symmetricKey
              BY <2>1, SymmetricKeyConstantLemma
            <6>2. QED
              BY <3>2, <6>1
              DEF newAuthenticator, newHistoryStateBinding, newStateHash,
                  newPrivateStateEnc, sResult, privateState
          (*****************************************************************)
          (* We can thus use the MACComplete property to show that the     *)
          (* generated MAC validates appropriately.  To do this, we first  *)
          (* need to prove some types.                                     *)
          (*****************************************************************)
          <5>2. LL1NVRAM.symmetricKey' \in SymmetricKeyType
            <6>1. LL1TypeInvariant'
              BY <2>1
            <6>2. QED
              BY <6>1, LL1SubtypeImplicationLemma DEF LL1SubtypeImplication
          <5>3. newHistoryStateBinding \in HashType
            BY <3>5
          (*****************************************************************)
          (* Then, we appeal to the MACComplete property in a              *)
          (* straightforward way.                                          *)
          (*****************************************************************)
          <5>4. QED
            BY <5>1, <5>2, <5>3, MACComplete
        (*******************************************************************)
        (* The new state hash defined by the LL1PerformOperation is in     *)
        (* HashType.                                                       *)
        (*******************************************************************)
        <4>2. newStateHash \in HashType
          BY <3>5
        (*******************************************************************)
        (* The new authenticator defined by the LL1PerformOperation *)
        (* action is in the primed set of observed authenticators,  *)
        (* as specified by the LL1PerformOperation action.                 *)
        (*******************************************************************)
        <4>3. newAuthenticator \in LL1ObservedAuthenticators'
          <5>1. LL1ObservedAuthenticators' =
              LL1ObservedAuthenticators \union {newAuthenticator}
            BY <3>2
            DEF newAuthenticator, newHistoryStateBinding, newStateHash,
                newHistorySummary, newPrivateStateEnc, sResult, privateState
          <5>2. QED
            BY <5>1
        (*******************************************************************)
        (* The history summary in the primed state of the NVRAM equals the  *)
        (* new history summary defined by the LL1PerformOperation.          *)
        (*******************************************************************)
        <4>4. LL1NVRAM.historySummary' = newHistorySummary 
          <5>1. LL1NVRAM' = [
                     historySummary |-> newHistorySummary,
                     symmetricKey |-> LL1NVRAM.symmetricKey ]
            BY <3>2 DEF newHistorySummary
          <5>2. QED
            BY <5>1
        (*******************************************************************)
        (* The above three conditions and the equality are sufficient to   *)
        (* satisfy the LL1NVRAMHistorySummaryUncorrupted predicate in the   *)
        (* primed state.                                                   *)
        (*******************************************************************)
        <4>5. QED
          BY <4>1, <4>2, <4>3, <4>4
          DEF LL1NVRAMHistorySummaryUncorrupted, LL1HistoryStateBindingAuthenticated,
              newHistoryStateBinding
      (*********************************************************************)
      (* The proof proper has two main steps.  The first main step is to   *)
      (* prove that the public and private states that are arguments to    *)
      (* the Service in LL1PerformOperation are identical to the      *)
      (* values of HLPublicState and HLPrivateState in the LL1Refinement.  *)
      (*********************************************************************)
      <3>8. /\  HLPublicState = LL1RAM.publicState
            /\  HLPrivateState = privateState
        (*******************************************************************)
        (* We show that the refined public state matches the public state  *)
        (* in the RAM and that the refined encrypted private state matches *)
        (* the encrypted private state in the RAM.  We will use the        *)
        (* HashCollisionResistant property.                                *)
        (*******************************************************************)
        <4>1. /\ HLPublicState = LL1RAM.publicState
              /\ refPrivateStateEnc = LL1RAM.privateStateEnc
          (*****************************************************************)
          (* We begin by proving some types we will need to appeal to the  *)
          (* HashCollisionResistant property.                              *)
          (*****************************************************************)
          <5>1. HLPublicState \in HashDomain
            <6>1. HLPublicState \in PublicStateType
              BY <2>1, <3>6 DEF LL1Refinement
            <6>2. QED
              BY <6>1 DEF HashDomain
          <5>2. /\  LL1RAM.publicState \in HashDomain
                /\  LL1RAM.privateStateEnc \in HashDomain
            <6>1. /\  LL1RAM.publicState \in PublicStateType
                  /\  LL1RAM.privateStateEnc \in PrivateStateEncType
              <7>1. LL1TypeInvariant
                BY <2>1
              <7>2. QED
                BY <7>1, LL1SubtypeImplicationLemma DEF LL1SubtypeImplication
            <6>2. QED
              BY <6>1 DEF HashDomain
          <5>3. refPrivateStateEnc \in HashDomain
            <6>1. refPrivateStateEnc \in PrivateStateEncType
              BY <3>3
            <6>2. QED
              BY <6>1 DEF HashDomain
          (*****************************************************************)
          (* We then prove that the refined state hash equals the state    *)
          (* hash defined by LL1PerformOperation.  We will use the         *)
          (* UniquenessInvariant.                                          *)
          (*****************************************************************)
          <5>4. refStateHash = stateHash
            (***************************************************************)
            (* We prove some types we will need for the                    *)
            (* UniquenessInvariant.                                        *)
            (***************************************************************)
            <6>1. refStateHash \in HashType
              BY <3>3
            <6>2. stateHash \in HashType
              BY <3>5
            (***************************************************************)
            (* We prove that the refined history state binding is           *)
            (* authenticated.  This follows directly from the              *)
            (* LL1Refinement.                                              *)
            (***************************************************************)
            <6>3. LL1HistoryStateBindingAuthenticated(refHistoryStateBinding)
              BY <2>1, <3>6 DEF LL1Refinement, refHistoryStateBinding, refStateHash, refPrivateStateEnc
            (***************************************************************)
            (* We prove that the history state binding defined by the       *)
            (* LL1PerformOperation action is authenticated.                *)
            (***************************************************************)
            <6>4. LL1HistoryStateBindingAuthenticated(historyStateBinding)
              (*************************************************************)
              (* The authenticator in the RAM authenticates the history    *)
              (* state binding, since this is an enablement condition of   *)
              (* the LL1PerformOperation action.                           *)
              (*************************************************************)
              <7>1. ValidateMAC(LL1NVRAM.symmetricKey, historyStateBinding, LL1RAM.authenticator)
                BY <3>2 DEF historyStateBinding, stateHash
              (*************************************************************)
              (* From the UnforgeabilityInvariant, the authenticator in    *)
              (* the RAM is in the set of observed authenticators.         *)
              (*************************************************************)
              <7>2. LL1RAM.authenticator \in LL1ObservedAuthenticators
                <8>1. UnforgeabilityInvariant
                  BY <2>1 DEF CorrectnessInvariants
                <8>2. historyStateBinding \in HashType
                  BY <3>5
                <8>3. QED
                  BY <7>1, <8>1, <8>2 DEF UnforgeabilityInvariant
              (*************************************************************)
              (* The above two conditions satisfy the definition of        *)
              (* LL1HistoryStateBindingAuthenticated.                       *)
              (*************************************************************)
              <7>3. QED
                BY <7>1, <7>2 DEF LL1HistoryStateBindingAuthenticated
            (***************************************************************)
            (* Since both history state bindings are authenticated, it      *)
            (* follows that the two state hashes are equal.                *)
            (***************************************************************)
            <6>5. QED
              <7>1. UniquenessInvariant
                BY <2>1 DEF CorrectnessInvariants
              <7>2. LL1NVRAM.historySummary = LL1RAM.historySummary
                BY <3>2
              <7>3. QED
                BY <7>1, <7>2, <6>1, <6>2, <6>3, <6>4
                DEF UniquenessInvariant, refHistoryStateBinding, historyStateBinding
          <5>5. QED
            BY <5>1, <5>2, <5>3, <5>4, HashCollisionResistant DEF refStateHash, stateHash
        (*******************************************************************)
        (* From the previous step, it immediately follows that the refined *)
        (* public state matches the public state in the RAM.               *)
        (*******************************************************************)
        <4>2. HLPublicState = LL1RAM.publicState
          BY <4>1
        (*******************************************************************)
        (* It is slightly more involved to show that the refined private   *)
        (* state matches the private state that is the decryption of the   *)
        (* encrypted private state in the RAM.  We need to employ the      *)
        (* correctness of the symmetric crypto operations.                 *)
        (*******************************************************************)
        <4>3. HLPrivateState = privateState
          <5>1. refPrivateStateEnc = LL1RAM.privateStateEnc
            BY <4>1
          <5>2. LL1NVRAM.symmetricKey \in SymmetricKeyType
            <6>1. LL1TypeInvariant
              BY <2>1
            <6>2. QED
              BY <6>1, LL1SubtypeImplicationLemma DEF LL1SubtypeImplication
          <5>3. HLPrivateState \in PrivateStateType
            BY <2>1, <3>6 DEF LL1Refinement
          <5>4. QED
            BY <5>1, <5>2, <5>3, SymmetricCryptoCorrect DEF refPrivateStateEnc, privateState
        <4>4. QED
          BY <4>2, <4>3
      (*********************************************************************)
      (* The second main step is to prove that the public and private      *)
      (* states that are produced by the Service in                   *)
      (* LL1PerformOperation are identical to the primed values of         *)
      (* HLPublicState and HLPrivateState in the LL1Refinement.            *)
      (*********************************************************************)
      <3>9. /\  HLPublicState' = sResult.newPublicState
            /\  HLPrivateState' = sResult.newPrivateState
        (*******************************************************************)
        (* We show that the primed refined public state matches the public *)
        (* state produced by the invocation of Service in             *)
        (* LL1PerformOperation, and that the primed refined encrypted      *)
        (* private state matches the encryption of the private state       *)
        (* produced by the invocation of Service in                   *)
        (* LL1PerformOperation.  We will use the HashCollisionResistant    *)
        (* property.                                                       *)
        (*******************************************************************)
        <4>1. /\ HLPublicState' = sResult.newPublicState
              /\ refPrivateStateEnc' = newPrivateStateEnc
          (*****************************************************************)
          (* We prove some types we will need for the                      *)
          (* HashCollisionResistant property.                              *)
          (*****************************************************************)
          <5>1. HLPublicState' \in HashDomain
            <6>1. HLPublicState' \in PublicStateType
              BY <2>1, <3>7 DEF LL1Refinement
            <6>2. QED
              BY <6>1 DEF HashDomain
          <5>2. refPrivateStateEnc' \in HashDomain 
            <6>1. refPrivateStateEnc' \in PrivateStateEncType
              BY <3>4
            <6>2. QED
              BY <6>1 DEF HashDomain
          <5>3. sResult.newPublicState \in HashDomain
            <6>1. sResult.newPublicState \in PublicStateType
              BY <3>5
            <6>2. QED
              BY <6>1 DEF HashDomain
          <5>4. newPrivateStateEnc \in HashDomain
            <6>1. newPrivateStateEnc \in PrivateStateEncType
              BY <3>5
            <6>2. QED
              BY <6>1 DEF HashDomain
          (*****************************************************************)
          (* We then prove that the primed refined state hash equals the   *)
          (* new state hash defined by LL1PerformOperation.  We will use   *)
          (* the UniquenessInvariant.                                      *)
          (*****************************************************************)
          <5>5. refStateHash' = newStateHash
            (***************************************************************)
            (* We prove that the refined history state binding is           *)
            (* authenticated in the primed state.  This follows directly   *)
            (* from the LL1Refinement.                                     *)
            (***************************************************************)
            <6>1. LL1HistoryStateBindingAuthenticated(refHistoryStateBinding)'
              BY <2>1, <3>7 DEF LL1Refinement, refHistoryStateBinding, refStateHash, refPrivateStateEnc
            (***************************************************************)
            (* Ideally, at this point we would prove that the new history  *)
            (* state binding defined by the LL1PerformOperation action is  *)
            (* authenticated in the primed state.  However, we cannot      *)
            (* prove this, because TLA+ does not allow us to prime an      *)
            (* operator, only an expression.  If we prime the entire       *)
            (* expression, the newHistoryStateBinding argument also        *)
            (* becomes primed, which changes its meaning.  So instead, we  *)
            (* merely prove that the new authenticator defined by the      *)
            (* LL1PerformOperation action authenticates the new history    *)
            (* state binding.  We will use the MACComplete property.       *)
            (***************************************************************)
            <6>2. ValidateMAC(LL1NVRAM.symmetricKey', newHistoryStateBinding, newAuthenticator)
              (*************************************************************)
              (* The new authenticator was genererated as a MAC of  *)
              (* the new history state binding by LL1PerformOperation,      *)
              (* using the unchanged symmetric key in the NVRAM.           *)
              (*************************************************************)
              <7>1. newAuthenticator = GenerateMAC(LL1NVRAM.symmetricKey', newHistoryStateBinding)
                <8>1. UNCHANGED LL1NVRAM.symmetricKey
                  BY <2>1, SymmetricKeyConstantLemma
                <8>2. QED
                  BY <3>2, <8>1
                  DEF newAuthenticator, newHistoryStateBinding, newStateHash,
                      newPrivateStateEnc, sResult, privateState
              (*************************************************************)
              (* We can thus use the MACComplete property to show that the *)
              (* generated MAC validates appropriately.  To do this, we    *)
              (* first need to prove some types.                           *)
              (*************************************************************)
              <7>2. LL1NVRAM.symmetricKey' \in SymmetricKeyType
                <8>1. LL1TypeInvariant'
                  BY <2>1
                <8>2. QED
                  BY <8>1, LL1SubtypeImplicationLemma DEF LL1SubtypeImplication
              <7>3. newHistoryStateBinding \in HashType
                BY <3>5
              (*************************************************************)
              (* Then, we appeal to the MACComplete property in a          *)
              (* straightforward way.                                      *)
              (*************************************************************)
              <7>4. QED
                BY <7>1, <7>2, <7>3, MACComplete
            (***************************************************************)
            (* Ideally, at this point we would prove that because both     *)
            (* history state bindings are authenticated, it follows that    *)
            (* the two state hashes are equal.  However, we have not       *)
            (* proven that the new state binding defined by the            *)
            (* LL1PerformOperation action is authenticated in the primed   *)
            (* state, because TLA+ does not allow us to prime an operator, *)
            (* only an expression.  So, we have to expand the definition   *)
            (* of LL1HistoryStateBindingAuthenticated for the steps of this *)
            (* proof.                                                      *)
            (***************************************************************)
            <6>3. QED
              (*************************************************************)
              (* This is the definition of the UniquenessInvariant, with   *)
              (* the LETs instantiated and                                 *)
              (* LL1HistoryStateBindingAuthenticated fully expanded.        *)
              (*************************************************************)
              <7>1. \A stateHash1, stateHash2 \in HashType :
                          ( /\  \E authenticator \in LL1ObservedAuthenticators' :
                                    ValidateMAC(
                                        LL1NVRAM.symmetricKey',
                                        Hash(LL1NVRAM.historySummary', stateHash1),
                                        authenticator)
                            /\  \E authenticator \in LL1ObservedAuthenticators' :
                                    ValidateMAC(
                                        LL1NVRAM.symmetricKey',
                                        Hash(LL1NVRAM.historySummary', stateHash2),
                                        authenticator) )
                        =>
                            stateHash1 = stateHash2
                <8>1. UniquenessInvariant'
                  BY <2>1 DEF CorrectnessInvariants
                <8>2. QED
                  BY <8>1 DEF UniquenessInvariant, LL1HistoryStateBindingAuthenticated
              (*************************************************************)
              (* The universal quantifiers in the previous step range over *)
              (* HashType, sowe need to prove that the state hashes are in *)
              (* HashType.                                                 *)
              (*************************************************************)
              <7>2. refStateHash' \in HashType
                BY <3>4
              <7>3. newStateHash \in HashType
                BY <3>5
              (*************************************************************)
              (* This is the first conjunct in the antecedent of the       *)
              (* implication in the expanded UniquenessInvariant.  It      *)
              (* follows directly from the fact that the refined history   *)
              (* state binding is authenticated in the primed state.       *)
              (*************************************************************)
              <7>4. \E authenticator \in LL1ObservedAuthenticators' :
                        ValidateMAC(
                            LL1NVRAM.symmetricKey',
                            Hash(LL1NVRAM.historySummary', refStateHash'),
                            authenticator)
                BY <6>1 DEF LL1HistoryStateBindingAuthenticated, refHistoryStateBinding
              (*************************************************************)
              (* This is the second conjunct in the antecedent of the      *)
              (* implication in the expanded UniquenessInvariant.  The new *)
              (* authenticator defined by the LL1PerformOperation   *)
              (* action will serve as a witness for the existential        *)
              (* quantifier.                                               *)
              (*************************************************************)
              <7>5. \E authenticator \in LL1ObservedAuthenticators' :
                        ValidateMAC(
                            LL1NVRAM.symmetricKey',
                            Hash(LL1NVRAM.historySummary', newStateHash),
                            authenticator)
                (***********************************************************)
                (* The new authenticator defined by the             *)
                (* LL1PerformOperation action is in the primed set of      *)
                (* observed authenticators, as specified by the     *)
                (* LL1PerformOperation action.                             *)
                (***********************************************************)
                <8>1. newAuthenticator \in LL1ObservedAuthenticators'
                  <9>1. LL1ObservedAuthenticators' =
                      LL1ObservedAuthenticators \union {newAuthenticator}
                    BY <3>2
                    DEF newAuthenticator, newHistoryStateBinding, newStateHash,
                        newHistorySummary, newPrivateStateEnc, sResult, privateState
                  <9>2. QED
                    BY <9>1
                (***********************************************************)
                (* The new authenticator defined by the                    *)
                (* LL1PerformOperation action authenticates the new        *)
                (* history state binding defined by this action.  Because  *)
                (* the history summary in the primed NVRAM equals the new  *)
                (* history summary defined by the action, we can derive an *)
                (* expression that satisfies the exsistential expression   *)
                (* above.                                                  *)
                (***********************************************************)
                <8>2. ValidateMAC(
                        LL1NVRAM.symmetricKey',
                        Hash(LL1NVRAM.historySummary', newStateHash),
                        newAuthenticator)
                  <9>1. newHistoryStateBinding = Hash(LL1NVRAM.historySummary', newStateHash)
                    <10>1. LL1NVRAM.historySummary' = newHistorySummary
                      <11>1. LL1NVRAM' = [
                                 historySummary |-> newHistorySummary,
                                 symmetricKey |-> LL1NVRAM.symmetricKey ]
                        BY <3>2 DEF newHistorySummary
                      <11>2. QED
                        BY <11>1
                    <10>2. QED
                      BY <10>1 DEF newHistoryStateBinding
                  <9>2. QED
                    BY <6>2, <9>1
                <8>3. QED
                  BY <8>1, <8>2
              (*************************************************************)
              (* With both conjuncts in the antecedent true, the           *)
              (* conclusion readily follows.                               *)
              (*************************************************************)
              <7>6. QED
                BY <7>1, <7>2, <7>3, <7>4, <7>5
          <5>6. QED
            (***************************************************************)
            (* Ideally, this QED step should just read:                    *)
            (*                                                             *)
            (*  BY <<5>>1, <<5>>2, <<5>>3, <<5>>4, <<5>>5, HashCollisionResistant *)
            (*                                                             *)
            (* However, the prover seems to get a little confused in this  *)
            (* instance.  We make life easier for the prover by definining *)
            (* some local variables and hiding their definitions before    *)
            (* appealing to the HashCollisionResistant assumption.         *)
            (***************************************************************)
            <6> h1a == HLPublicState'
            <6> h2a == refPrivateStateEnc'
            <6> h1b == sResult.newPublicState
            <6> h2b == newPrivateStateEnc
            <6>1. h1a \in HashDomain
              BY <5>1
            <6>2. h2a \in HashDomain
              BY <5>2
            <6>3. h1b \in HashDomain
              BY <5>3
            <6>4. h2b \in HashDomain
              BY <5>4
            <6>5. Hash(h1a, h2a) = Hash(h1b, h2b)
              BY <5>5 DEF refStateHash, newStateHash
            <6>6. h1a = h1b /\ h2a = h2b
              <7> HIDE DEF h1a, h2a, h1b, h2b
              <7>1. QED
                BY <6>1, <6>2, <6>3, <6>4, <6>5, HashCollisionResistant
            <6>7. QED
              BY <6>6
        (*******************************************************************)
        (* To show that the primed refined high-level private state        *)
        (* matches the private state that is produced by the Service  *)
        (* invocation in LL1PerformOperation, we need to appeal to the     *)
        (* correctness of the symmetric crypto operations and the          *)
        (* SymmetricKeyConstantLemma.                                      *)
        (*******************************************************************)
        <4>2. HLPrivateState' = sResult.newPrivateState
          <5>1. SymmetricDecrypt(LL1NVRAM.symmetricKey', refPrivateStateEnc') =
                SymmetricDecrypt(LL1NVRAM.symmetricKey, newPrivateStateEnc)
            <6>1. refPrivateStateEnc' = newPrivateStateEnc
              BY <4>1
           <6>2. LL1NVRAM.symmetricKey' = LL1NVRAM.symmetricKey
              BY <2>1, SymmetricKeyConstantLemma
            <6>20. QED
              BY <6>1, <6>2
          <5>2. HLPrivateState' = SymmetricDecrypt(LL1NVRAM.symmetricKey', refPrivateStateEnc')
            <6>1. LL1NVRAM.symmetricKey' \in SymmetricKeyType
              <7>1. LL1TypeInvariant'
                BY <2>1
              <7>2. QED
                BY <7>1, LL1SubtypeImplicationLemma DEF LL1SubtypeImplication
            <6>2. HLPrivateState' \in PrivateStateType
              BY <2>1, <3>7 DEF LL1Refinement
            <6>3. QED
              BY <6>1, <6>2, SymmetricCryptoCorrect DEF refPrivateStateEnc
          <5>3. sResult.newPrivateState = SymmetricDecrypt(LL1NVRAM.symmetricKey, newPrivateStateEnc)
            <6>1. LL1NVRAM.symmetricKey  \in SymmetricKeyType
              <7>1. LL1TypeInvariant
                BY <2>1
              <7>2. QED
                BY <7>1, LL1SubtypeImplicationLemma DEF LL1SubtypeImplication
            <6>2. sResult.newPrivateState \in PrivateStateType
              BY <3>5
            <6>3. QED
              BY <6>1, <6>2, SymmetricCryptoCorrect DEF newPrivateStateEnc
          <5>4. QED
            BY <5>1, <5>2, <5>3
        <4>3. QED
          BY <4>1, <4>2
      (*********************************************************************)
      (* The QED step states each conjunct of the HLAdvanceService    *)
      (* action, which all follow directly from the two main steps above.  *)
      (*********************************************************************)
      <3>10. QED
        <4>1. input \in HLAvailableInputs
          <5>1. HLAvailableInputs = LL1AvailableInputs
            BY <2>1 DEF LL1Refinement
          <5>2. QED
            BY <5>1, <3>2
        <4>2. HLAlive = TRUE
          BY <2>1, <3>6 DEF LL1Refinement
        <4>3. HLPublicState' = hlSResult.newPublicState
          <5>1. HLPublicState' = sResult.newPublicState
            BY <3>9
          <5>2. sResult.newPublicState = hlSResult.newPublicState
            <6>1. sResult = hlSResult
              BY <3>8 DEF sResult, hlSResult
            <6>2. QED
              BY <6>1
          <5>3. QED
            BY <5>1, <5>2
        <4>4. HLPrivateState' = hlSResult.newPrivateState
          <5>1. HLPrivateState' = sResult.newPrivateState
            BY <3>9
          <5>2. sResult.newPrivateState = hlSResult.newPrivateState
            <6>1. sResult = hlSResult
              BY <3>8 DEF sResult, hlSResult
            <6>2. QED
              BY <6>1
          <5>3. QED
            BY <5>1, <5>2
        <4>5. HLObservedOutputs' = HLObservedOutputs \union {hlSResult.output}
          <5>1. LL1ObservedOutputs' = LL1ObservedOutputs \union {sResult.output}
            BY <3>2 DEF sResult, privateState
          <5>2. HLObservedOutputs = LL1ObservedOutputs
            BY <2>1 DEF LL1Refinement
          <5>3. HLObservedOutputs' = LL1ObservedOutputs'
            BY <2>1 DEF LL1Refinement
          <5>4. sResult.output = hlSResult.output
            <6>1. sResult = hlSResult
              BY <3>8 DEF sResult, hlSResult
            <6>2. QED
              BY <6>1
          <5>5. QED
            BY <5>1, <5>2, <5>3, <5>4
        <4>6. UNCHANGED HLAvailableInputs
          <5>1. UNCHANGED LL1AvailableInputs
            BY <3>2
          <5>2. /\ HLAvailableInputs = LL1AvailableInputs
                /\ HLAvailableInputs' = LL1AvailableInputs'
            BY <2>1 DEF LL1Refinement
          <5>3. QED
            BY <5>1, <5>2
        <4>7. UNCHANGED HLAlive
          BY <2>1, <3>6, <3>7, LL1NVRAMHistorySummaryUncorruptedEqualsHLAliveLemma
        <4>8. QED
          BY <4>1, <4>2, <4>3, <4>4, <4>5, <4>6, <4>7 DEF HLAdvanceService, hlSResult
    (***********************************************************************)
    (* A `^\basicname^' LL1RepeatOperation action refines to a high-level  *)
    (* stuttering step.                                                    *)
    (***********************************************************************)
    <2>5. LL1RepeatOperation => UNCHANGED HLVars
      <3>1. HAVE LL1RepeatOperation
      <3>2. PICK input \in LL1AvailableInputs : LL1RepeatOperation!(input)!1 
        BY <3>1 DEF LL1RepeatOperation
      (*********************************************************************)
      (* The HLAlive predicate is unchanged because the                    *)
      (* LL1NVRAMHistorySummaryUncorrupted predicate is unchanged.          *)
      (*********************************************************************)
      <3>3. UNCHANGED HLAlive
        (*******************************************************************)
        (* The LL1NVRAMHistorySummaryUncorruptedEqualsHLAliveLemma tells us *)
        (* that LL1NVRAMHistorySummaryUncorrupted and HLAlive are equal.    *)
        (*******************************************************************)
        <4>1. LL1NVRAMHistorySummaryUncorrupted = HLAlive
          <5>1. LL1Refinement
            BY <2>1
          <5>2. QED
            BY <5>1, LL1NVRAMHistorySummaryUncorruptedEqualsHLAliveLemma
        <4>2. LL1NVRAMHistorySummaryUncorrupted' = HLAlive'
          <5>1. LL1Refinement'
            BY <2>1
          <5>2. QED
            BY <5>1, LL1NVRAMHistorySummaryUncorruptedEqualsHLAliveLemma
        (*******************************************************************)
        (* Then, the LL1NVRAMHistorySummaryUncorruptedUnchangedLemma tells  *)
        (* us that the LL1NVRAMHistorySummaryUncorrupted predicate is       *)
        (* unchanged.                                                      *)
        (*******************************************************************)
        <4>3. UNCHANGED LL1NVRAMHistorySummaryUncorrupted
          <5>1. LL1TypeInvariant
            BY <2>1
          (*****************************************************************)
          (* The NVRAM is unchanged by definition of the                   *)
          (* LL1RepeatOperation action.                                    *)
          (*****************************************************************)
          <5>2. UNCHANGED LL1NVRAM
            BY <3>2
          (*****************************************************************)
          (* The                                                           *)
          (* LL1RepeatOperationUnchangedAuthenticatedHistoryStateBindingsLemma *)
          (* tells us that there is no change to the set of history state  *)
          (* bindings that have authenticators in the set                  *)
          (* LL1ObservedAuthenticators.                                    *)
          (*****************************************************************)
          <5>3. \A historyStateBinding \in HashType :
                    UNCHANGED LL1HistoryStateBindingAuthenticated(historyStateBinding)
            <6>1. LL1TypeInvariant /\ UnforgeabilityInvariant /\ InclusionInvariant
              BY <2>1 DEF CorrectnessInvariants
            <6>2. QED
              BY <3>1, <6>1, LL1RepeatOperationUnchangedAuthenticatedHistoryStateBindingsLemma
          <5>4. QED
            BY <5>1, <5>2, <5>3, LL1NVRAMHistorySummaryUncorruptedUnchangedLemma
        <4>4. QED
          BY <4>1, <4>2, <4>3
      (*********************************************************************)
      (* The mapping from LL1AvailableInputs to HLAvailableInputs is       *)
      (* direct.                                                           *)
      (*********************************************************************)
      <3>4. UNCHANGED HLAvailableInputs
        <4>1. UNCHANGED LL1AvailableInputs
          BY <3>2
        <4>2. /\ HLAvailableInputs = LL1AvailableInputs
              /\ HLAvailableInputs' = LL1AvailableInputs'
          BY <2>1 DEF LL1Refinement
        <4>3. QED
          BY <4>1, <4>2
      (*********************************************************************)
      (* The mapping from LL1ObservedOutputs to HLObservedOutputs is       *)
      (* direct.                                                           *)
      (*********************************************************************)
      <3>5. UNCHANGED HLObservedOutputs
        <4>1. UNCHANGED LL1ObservedOutputs
          <5>1. LL1TypeInvariant /\ UnforgeabilityInvariant /\ InclusionInvariant
            BY <2>1 DEF CorrectnessInvariants
          <5>2. QED
            BY <5>1, <3>1, LL1RepeatOperationUnchangedObseredOutputsLemma
        <4>2. /\ HLObservedOutputs = LL1ObservedOutputs
              /\ HLObservedOutputs' = LL1ObservedOutputs'
          BY <2>1 DEF LL1Refinement
        <4>3. QED
          BY <4>1, <4>2
      (*********************************************************************)
      (* We prove the stuttering of the high-level public and private      *)
      (* state by using the NonAdvancementLemma.                           *)
      (*********************************************************************)
      <3>6. UNCHANGED <<HLPublicState, HLPrivateState>>
        (*******************************************************************)
        (* Many of the antecedents for the NonAdvancementLemma come        *)
        (* directly from antecedents in the induction.                     *)
        (*******************************************************************)
        <4>1. /\  LL1Refinement
              /\  LL1Refinement'
              /\  LL1TypeInvariant
              /\  LL1TypeInvariant'
              /\  UniquenessInvariant
          BY <2>1 DEF CorrectnessInvariants
        (*******************************************************************)
        (* The NVRAM is unchanged by definition of the LL1RepeatOperation  *)
        (* action.                                                         *)
        (*******************************************************************)
        <4>2. UNCHANGED LL1NVRAM
          BY <3>2
        (*******************************************************************)
        (* The                                                             *)
        (* LL1RepeatOperationUnchangedAuthenticatedHistoryStateBindingsLemma *)
        (* tells us that there is no change to the set of history state    *)
        (* bindings that have authenticators in the set                    *)
        (* LL1ObservedAuthenticators.                                      *)
        (*******************************************************************)
        <4>3. \A historyStateBinding \in HashType :
                  UNCHANGED LL1HistoryStateBindingAuthenticated(historyStateBinding)
          <5>1. LL1TypeInvariant /\ UnforgeabilityInvariant /\ InclusionInvariant
            BY <2>1 DEF CorrectnessInvariants
          <5>2. QED
            BY <5>1, <3>1, LL1RepeatOperationUnchangedAuthenticatedHistoryStateBindingsLemma
        (*******************************************************************)
        (* We have all of the antecedents for the NonAdvancementLemma, so  *)
        (* we can apply it directly.                                       *)
        (*******************************************************************)
        <4>4. QED
            BY <4>1, <4>2, <4>3, NonAdvancementLemma
      <3>7. QED
        BY <3>3, <3>4, <3>5, <3>6 DEF HLVars
    (***********************************************************************)
    (* A `^\basicname^' LL1Restart action refines to a high-level          *)
    (* stuttering step.                                                    *)
    (***********************************************************************)
    <2>6. LL1Restart => UNCHANGED HLVars
      <3>1. HAVE LL1Restart
      <3>2. PICK untrustedStorage \in LL1UntrustedStorageType,
                 randomSymmetricKey \in SymmetricKeyType \ {LL1NVRAM.symmetricKey},
                 hash \in HashType :
                 LL1Restart!(untrustedStorage, randomSymmetricKey, hash)
        BY <3>1 DEF LL1Restart
      (*********************************************************************)
      (* The HLAlive predicate is unchanged because the                    *)
      (* LL1NVRAMHistorySummaryUncorrupted predicate is unchanged.          *)
      (*********************************************************************)
      <3>3. UNCHANGED HLAlive
        (*******************************************************************)
        (* The LL1NVRAMHistorySummaryUncorruptedEqualsHLAliveLemma tells us *)
        (* that LL1NVRAMHistorySummaryUncorrupted and HLAlive are equal.    *)
        (*******************************************************************)
        <4>1. LL1NVRAMHistorySummaryUncorrupted = HLAlive
          <5>1. LL1Refinement
            BY <2>1
          <5>2. QED
            BY <5>1, LL1NVRAMHistorySummaryUncorruptedEqualsHLAliveLemma
        <4>2. LL1NVRAMHistorySummaryUncorrupted' = HLAlive'
          <5>1. LL1Refinement'
            BY <2>1
          <5>2. QED
            BY <5>1, LL1NVRAMHistorySummaryUncorruptedEqualsHLAliveLemma
        (*******************************************************************)
        (* Then, the LL1NVRAMHistorySummaryUncorruptedUnchangedLemma tells  *)
        (* us that the LL1NVRAMHistorySummaryUncorrupted predicate is       *)
        (* unchanged.                                                      *)
        (*******************************************************************)
        <4>3. UNCHANGED LL1NVRAMHistorySummaryUncorrupted
          <5>1. LL1TypeInvariant
            BY <2>1
          (*****************************************************************)
          (* The NVRAM is unchanged by definition of the LL1Restart        *)
          (* action.                                                       *)
          (*****************************************************************)
          <5>2. UNCHANGED LL1NVRAM
            BY <3>2
          (*****************************************************************)
          (* The UnchangedAuthenticatedHistoryStateBindingsLemma tells us   *)
          (* that there is no change to the set of history state bindings   *)
          (* that have authenticators in the set                          *)
          (* LL1ObservedAuthenticators.                              *)
          (*****************************************************************)
          <5>3. \A historyStateBinding \in HashType :
                    UNCHANGED LL1HistoryStateBindingAuthenticated(historyStateBinding)
            <6>1. UNCHANGED <<LL1NVRAM, LL1ObservedAuthenticators>>
              BY <3>2
            <6>2. QED
              BY <6>1, UnchangedAuthenticatedHistoryStateBindingsLemma
          (*****************************************************************)
          (* We have all of the antecedents for the                        *)
          (* LL1NVRAMHistorySummaryUncorruptedUnchangedLemma, so we can     *)
          (* apply it directly.                                            *)
          (*****************************************************************)
          <5>4. QED
            BY <5>1, <5>2, <5>3, LL1NVRAMHistorySummaryUncorruptedUnchangedLemma
        <4>4. QED
          BY <4>1, <4>2, <4>3
      (*********************************************************************)
      (* The mapping from LL1AvailableInputs to HLAvailableInputs is       *)
      (* direct.                                                           *)
      (*********************************************************************)
      <3>4. UNCHANGED HLAvailableInputs
        <4>1. UNCHANGED LL1AvailableInputs
          BY <3>2
        <4>2. /\ HLAvailableInputs = LL1AvailableInputs
              /\ HLAvailableInputs' = LL1AvailableInputs'
          BY <2>1 DEF LL1Refinement
        <4>3. QED
          BY <4>1, <4>2
      (*********************************************************************)
      (* The mapping from LL1ObservedOutputs to HLObservedOutputs is       *)
      (* direct.                                                           *)
      (*********************************************************************)
      <3>5. UNCHANGED HLObservedOutputs
        <4>1. UNCHANGED LL1ObservedOutputs
          BY <3>2
        <4>2. /\ HLObservedOutputs = LL1ObservedOutputs
              /\ HLObservedOutputs' = LL1ObservedOutputs'
          BY <2>1 DEF LL1Refinement
        <4>3. QED
          BY <4>1, <4>2
      (*********************************************************************)
      (* We prove the stuttering of the high-level public and private      *)
      (* state by using the NonAdvancementLemma.                           *)
      (*********************************************************************)
      <3>6. UNCHANGED <<HLPublicState, HLPrivateState>>
        (*******************************************************************)
        (* Many of the antecedents for the NonAdvancementLemma come        *)
        (* directly from antecedents in the induction.                     *)
        (*******************************************************************)
        <4>1. /\  LL1Refinement
              /\  LL1Refinement'
              /\  LL1TypeInvariant
              /\  LL1TypeInvariant'
              /\  UniquenessInvariant
          BY <2>1 DEF CorrectnessInvariants
        (*******************************************************************)
        (* The LL1NVRAM is unchanged by definition of the LL1Restart       *)
        (* action.                                                         *)
        (*******************************************************************)
        <4>2. UNCHANGED LL1NVRAM
          BY <3>2
        (*******************************************************************)
        (* The UnchangedAuthenticatedHistoryStateBindingsLemma tells us     *)
        (* that there is no change to the set of history state bindings     *)
        (* that have authenticators in the set                            *)
        (* LL1ObservedAuthenticators.                                *)
        (*******************************************************************)
        <4>3. \A historyStateBinding \in HashType :
                  UNCHANGED LL1HistoryStateBindingAuthenticated(historyStateBinding)
          <5>1. UNCHANGED <<LL1NVRAM, LL1ObservedAuthenticators>>
            BY <3>2
          <5>2. QED
            BY <5>1, UnchangedAuthenticatedHistoryStateBindingsLemma
        (*******************************************************************)
        (* We have all of the antecedents for the NonAdvancementLemma, so  *)
        (* we can apply it directly.                                       *)
        (*******************************************************************)
        <4>4. QED
          BY <4>1, <4>2, <4>3, NonAdvancementLemma
      <3>7. QED
        BY <3>3, <3>4, <3>5, <3>6 DEF HLVars
    (***********************************************************************)
    (* A `^\basicname^' LL1ReadDisk action refines to a high-level         *)
    (* stuttering step.                                                    *)
    (***********************************************************************)
    <2>7. LL1ReadDisk => UNCHANGED HLVars
      <3>1. HAVE LL1ReadDisk
      (*********************************************************************)
      (* The HLAlive predicate is unchanged because the                    *)
      (* LL1NVRAMHistorySummaryUncorrupted predicate is unchanged.          *)
      (*********************************************************************)
      <3>2. UNCHANGED HLAlive
        (*******************************************************************)
        (* The LL1NVRAMHistorySummaryUncorruptedEqualsHLAliveLemma tells us *)
        (* that LL1NVRAMHistorySummaryUncorrupted and HLAlive are equal.    *)
        (*******************************************************************)
        <4>1. LL1NVRAMHistorySummaryUncorrupted = HLAlive
          <5>1. LL1Refinement
            BY <2>1
          <5>2. QED
            BY <5>1, LL1NVRAMHistorySummaryUncorruptedEqualsHLAliveLemma
        <4>2. LL1NVRAMHistorySummaryUncorrupted' = HLAlive'
          <5>1. LL1Refinement'
            BY <2>1
          <5>2. QED
            BY <5>1, LL1NVRAMHistorySummaryUncorruptedEqualsHLAliveLemma
        (*******************************************************************)
        (* Then, the LL1NVRAMHistorySummaryUncorruptedUnchangedLemma tells  *)
        (* us that the LL1NVRAMHistorySummaryUncorrupted predicate is       *)
        (* unchanged.                                                      *)
        (*******************************************************************)
        <4>3. UNCHANGED LL1NVRAMHistorySummaryUncorrupted
          <5>1. LL1TypeInvariant
            BY <2>1
          (*****************************************************************)
          (* The NVRAM is unchanged by definition of the LL1ReadDisk       *)
          (* action.                                                       *)
          (*****************************************************************)
          <5>2. UNCHANGED LL1NVRAM
            BY <3>1 DEF LL1ReadDisk
          (*****************************************************************)
          (* The UnchangedAuthenticatedHistoryStateBindingsLemma tells us   *)
          (* that there is no change to the set of history state bindings   *)
          (* that have authenticators in the set                          *)
          (* LL1ObservedAuthenticators.                              *)
          (*****************************************************************)
          <5>3. \A historyStateBinding \in HashType :
                    UNCHANGED LL1HistoryStateBindingAuthenticated(historyStateBinding)
            <6>1. UNCHANGED <<LL1NVRAM, LL1ObservedAuthenticators>>
              BY <3>1 DEF LL1ReadDisk
            <6>2. QED
              BY <6>1, UnchangedAuthenticatedHistoryStateBindingsLemma
          (*****************************************************************)
          (* We have all of the antecedents for the                        *)
          (* LL1NVRAMHistorySummaryUncorruptedUnchangedLemma, so we can     *)
          (* apply it directly.                                            *)
          (*****************************************************************)
          <5>4. QED
            BY <5>1, <5>2, <5>3, LL1NVRAMHistorySummaryUncorruptedUnchangedLemma
        <4>4. QED
          BY <4>1, <4>2, <4>3
      (*********************************************************************)
      (* The mapping from LL1AvailableInputs to HLAvailableInputs is       *)
      (* direct.                                                           *)
      (*********************************************************************)
      <3>3. UNCHANGED HLAvailableInputs
        <4>1. UNCHANGED LL1AvailableInputs
          BY <3>1 DEF LL1ReadDisk
        <4>2. /\ HLAvailableInputs = LL1AvailableInputs
              /\ HLAvailableInputs' = LL1AvailableInputs'
          BY <2>1 DEF LL1Refinement
        <4>3. QED
          BY <4>1, <4>2
      (*********************************************************************)
      (* The mapping from LL1ObservedOutputs to HLObservedOutputs is       *)
      (* direct.                                                           *)
      (*********************************************************************)
      <3>4. UNCHANGED HLObservedOutputs
        <4>1. UNCHANGED LL1ObservedOutputs
          BY <3>1 DEF LL1ReadDisk
        <4>2. /\ HLObservedOutputs = LL1ObservedOutputs
              /\ HLObservedOutputs' = LL1ObservedOutputs'
          BY <2>1 DEF LL1Refinement
        <4>3. QED
          BY <4>1, <4>2
      (*********************************************************************)
      (* We prove the stuttering of the high-level public and private      *)
      (* state by using the NonAdvancementLemma.                           *)
      (*********************************************************************)
      <3>5. UNCHANGED <<HLPublicState, HLPrivateState>>
        (*******************************************************************)
        (* Many of the antecedents for the NonAdvancementLemma come        *)
        (* directly from antecedents in the induction.                     *)
        (*******************************************************************)
        <4>1. /\  LL1Refinement
              /\  LL1Refinement'
              /\  LL1TypeInvariant
              /\  LL1TypeInvariant'
              /\  UniquenessInvariant
          BY <2>1 DEF CorrectnessInvariants
        (*******************************************************************)
        (* The NVRAM is unchanged by definition of the LL1ReadDisk action. *)
        (*******************************************************************)
        <4>2. UNCHANGED LL1NVRAM
          BY <3>1 DEF LL1ReadDisk
        (*******************************************************************)
        (* The UnchangedAuthenticatedHistoryStateBindingsLemma tells us     *)
        (* that there is no change to the set of history state bindings     *)
        (* that have authenticators in the set                            *)
        (* LL1ObservedAuthenticators.                                *)
        (*******************************************************************)
        <4>3. \A historyStateBinding \in HashType :
                  UNCHANGED LL1HistoryStateBindingAuthenticated(historyStateBinding)
          <5>1. UNCHANGED <<LL1NVRAM, LL1ObservedAuthenticators>>
            BY <3>1 DEF LL1ReadDisk
          <5>2. QED
            BY <5>1, UnchangedAuthenticatedHistoryStateBindingsLemma
        (*******************************************************************)
        (* We have all of the antecedents for the NonAdvancementLemma, so  *)
        (* we can apply it directly.                                       *)
        (*******************************************************************)
        <4>4. QED
          BY <4>1, <4>2, <4>3, NonAdvancementLemma
      <3>6. QED
        BY <3>2, <3>3, <3>4, <3>5 DEF HLVars
    (***********************************************************************)
    (* A `^\basicname^' LL1WriteDisk action refines to a high-level        *)
    (* stuttering step.                                                    *)
    (***********************************************************************)
    <2>8. LL1WriteDisk => UNCHANGED HLVars
      <3>1. HAVE LL1WriteDisk
      (*********************************************************************)
      (* The HLAlive predicate is unchanged because the                    *)
      (* LL1NVRAMHistorySummaryUncorrupted predicate is unchanged.          *)
      (*********************************************************************)
      <3>2. UNCHANGED HLAlive
        (*******************************************************************)
        (* The LL1NVRAMHistorySummaryUncorruptedEqualsHLAliveLemma tells us *)
        (* that LL1NVRAMHistorySummaryUncorrupted and HLAlive are equal.    *)
        (*******************************************************************)
        <4>1. LL1NVRAMHistorySummaryUncorrupted = HLAlive
          <5>1. LL1Refinement
            BY <2>1
          <5>2. QED
            BY <5>1, LL1NVRAMHistorySummaryUncorruptedEqualsHLAliveLemma
        <4>2. LL1NVRAMHistorySummaryUncorrupted' = HLAlive'
          <5>1. LL1Refinement'
            BY <2>1
          <5>2. QED
            BY <5>1, LL1NVRAMHistorySummaryUncorruptedEqualsHLAliveLemma
        (*******************************************************************)
        (* Then, the LL1NVRAMHistorySummaryUncorruptedUnchangedLemma tells  *)
        (* us that the LL1NVRAMHistorySummaryUncorrupted predicate is       *)
        (* unchanged.                                                      *)
        (*******************************************************************)
        <4>3. UNCHANGED LL1NVRAMHistorySummaryUncorrupted
          <5>1. LL1TypeInvariant
            BY <2>1
          (*****************************************************************)
          (* The NVRAM is unchanged by definition of the LL1WriteDisk      *)
          (* action.                                                       *)
          (*****************************************************************)
          <5>2. UNCHANGED LL1NVRAM
            BY <3>1 DEF LL1WriteDisk
          (*****************************************************************)
          (* The UnchangedAuthenticatedHistoryStateBindingsLemma tells us   *)
          (* that there is no change to the set of history state bindings   *)
          (* that have authenticators in the set                          *)
          (* LL1ObservedAuthenticators.                              *)
          (*****************************************************************)
          <5>3. \A historyStateBinding \in HashType :
                    UNCHANGED LL1HistoryStateBindingAuthenticated(historyStateBinding)
            <6>1. UNCHANGED <<LL1NVRAM, LL1ObservedAuthenticators>>
              BY <3>1 DEF LL1WriteDisk
            <6>2. QED
              BY <6>1, UnchangedAuthenticatedHistoryStateBindingsLemma
          (*****************************************************************)
          (* We have all of the antecedents for the                        *)
          (* LL1NVRAMHistorySummaryUncorruptedUnchangedLemma, so we can     *)
          (* apply it directly.                                            *)
          (*****************************************************************)
          <5>4. QED
            BY <5>1, <5>2, <5>3, LL1NVRAMHistorySummaryUncorruptedUnchangedLemma
        <4>4. QED
          BY <4>1, <4>2, <4>3
      (*********************************************************************)
      (* The mapping from LL1AvailableInputs to HLAvailableInputs is       *)
      (* direct.                                                           *)
      (*********************************************************************)
      <3>3. UNCHANGED HLAvailableInputs
        <4>1. UNCHANGED LL1AvailableInputs
          BY <3>1 DEF LL1WriteDisk
        <4>2. /\ HLAvailableInputs = LL1AvailableInputs
              /\ HLAvailableInputs' = LL1AvailableInputs'
          BY <2>1 DEF LL1Refinement
        <4>3. QED
          BY <4>1, <4>2
      (*********************************************************************)
      (* The mapping from LL1ObservedOutputs to HLObservedOutputs is       *)
      (* direct.                                                           *)
      (*********************************************************************)
      <3>4. UNCHANGED HLObservedOutputs
        <4>1. UNCHANGED LL1ObservedOutputs
          BY <3>1 DEF LL1WriteDisk
        <4>2. /\ HLObservedOutputs = LL1ObservedOutputs
              /\ HLObservedOutputs' = LL1ObservedOutputs'
          BY <2>1 DEF LL1Refinement
        <4>3. QED
          BY <4>1, <4>2
      (*********************************************************************)
      (* We prove the stuttering of the high-level public and private      *)
      (* state by using the NonAdvancementLemma.                           *)
      (*********************************************************************)
      <3>5. UNCHANGED <<HLPublicState, HLPrivateState>>
        (*******************************************************************)
        (* Many of the antecedents for the NonAdvancementLemma come        *)
        (* directly from antecedents in the induction.                     *)
        (*******************************************************************)
        <4>1. /\  LL1Refinement
              /\  LL1Refinement'
              /\  LL1TypeInvariant
              /\  LL1TypeInvariant'
              /\  UniquenessInvariant
          BY <2>1 DEF CorrectnessInvariants
        (*******************************************************************)
        (* The NVRAM is unchanged by definition of the LL1WriteDisk        *)
        (* action.                                                         *)
        (*******************************************************************)
        <4>2. UNCHANGED LL1NVRAM
          BY <3>1 DEF LL1WriteDisk
        (*******************************************************************)
        (* The UnchangedAuthenticatedHistoryStateBindingsLemma tells us     *)
        (* that there is no change to the set of history state bindings     *)
        (* that have authenticators in the set                            *)
        (* LL1ObservedAuthenticators.                                *)
        (*******************************************************************)
        <4>3. \A historyStateBinding \in HashType :
                  UNCHANGED LL1HistoryStateBindingAuthenticated(historyStateBinding)
          <5>1. UNCHANGED <<LL1NVRAM, LL1ObservedAuthenticators>>
            BY <3>1 DEF LL1WriteDisk
          <5>2. QED
            BY <5>1, UnchangedAuthenticatedHistoryStateBindingsLemma
        (*******************************************************************)
        (* We have all of the antecedents for the NonAdvancementLemma, so  *)
        (* we can apply it directly.                                       *)
        (*******************************************************************)
        <4>4. QED
          BY <4>1, <4>2, <4>3, NonAdvancementLemma
      <3>6. QED
        BY <3>2, <3>3, <3>4, <3>5 DEF HLVars
    (***********************************************************************)
    (* A `^\basicname^' LL1CorruptRAM action refines to a high-level       *)
    (* stuttering step.                                                    *)
    (***********************************************************************)
    <2>9. LL1CorruptRAM => UNCHANGED HLVars
      <3>1. HAVE LL1CorruptRAM
      <3>2. PICK untrustedStorage \in LL1UntrustedStorageType,
                 fakeSymmetricKey \in SymmetricKeyType \ {LL1NVRAM.symmetricKey},
                 hash \in HashType :
                 LL1CorruptRAM!(untrustedStorage, fakeSymmetricKey, hash)
        BY <3>1 DEF LL1CorruptRAM
      (*********************************************************************)
      (* The HLAlive predicate is unchanged because the                    *)
      (* LL1NVRAMHistorySummaryUncorrupted predicate is unchanged.          *)
      (*********************************************************************)
      <3>3. UNCHANGED HLAlive
        (*******************************************************************)
        (* The LL1NVRAMHistorySummaryUncorruptedEqualsHLAliveLemma tells us *)
        (* that LL1NVRAMHistorySummaryUncorrupted and HLAlive are equal.    *)
        (*******************************************************************)
        <4>1. LL1NVRAMHistorySummaryUncorrupted = HLAlive
          <5>1. LL1Refinement
            BY <2>1
          <5>2. QED
            BY <5>1, LL1NVRAMHistorySummaryUncorruptedEqualsHLAliveLemma
        <4>2. LL1NVRAMHistorySummaryUncorrupted' = HLAlive'
          <5>1. LL1Refinement'
            BY <2>1
          <5>2. QED
            BY <5>1, LL1NVRAMHistorySummaryUncorruptedEqualsHLAliveLemma
        (*******************************************************************)
        (* Then, the LL1NVRAMHistorySummaryUncorruptedUnchangedLemma tells  *)
        (* us that the LL1NVRAMHistorySummaryUncorrupted predicate is       *)
        (* unchanged.                                                      *)
        (*******************************************************************)
        <4>3. UNCHANGED LL1NVRAMHistorySummaryUncorrupted
          <5>1. LL1TypeInvariant
            BY <2>1
          (*****************************************************************)
          (* The LL1NVRAM is unchanged by definition of the LL1CorruptRAM  *)
          (* action.                                                       *)
          (*****************************************************************)
          <5>2. UNCHANGED LL1NVRAM
            BY <3>2
          (*****************************************************************)
          (* The UnchangedAuthenticatedHistoryStateBindingsLemma tells us   *)
          (* that there is no change to the set of history state bindings   *)
          (* that have authenticators in the set                          *)
          (* LL1ObservedAuthenticators.                              *)
          (*****************************************************************)
          <5>3. \A historyStateBinding \in HashType :
                    UNCHANGED LL1HistoryStateBindingAuthenticated(historyStateBinding)
            <6>1. UNCHANGED <<LL1NVRAM, LL1ObservedAuthenticators>>
              BY <3>2
            <6>2. QED
              BY <6>1, UnchangedAuthenticatedHistoryStateBindingsLemma
          (*****************************************************************)
          (* We have all of the antecedents for the                        *)
          (* LL1NVRAMHistorySummaryUncorruptedUnchangedLemma, so we can     *)
          (* apply it directly.                                            *)
          (*****************************************************************)
          <5>4. QED
            BY <5>1, <5>2, <5>3, LL1NVRAMHistorySummaryUncorruptedUnchangedLemma
        <4>4. QED
          BY <4>1, <4>2, <4>3
      (*********************************************************************)
      (* The mapping from LL1AvailableInputs to HLAvailableInputs is       *)
      (* direct.                                                           *)
      (*********************************************************************)
      <3>4. UNCHANGED HLAvailableInputs
        <4>1. UNCHANGED LL1AvailableInputs
          BY <3>2
        <4>2. /\ HLAvailableInputs = LL1AvailableInputs
              /\ HLAvailableInputs' = LL1AvailableInputs'
          BY <2>1 DEF LL1Refinement
        <4>3. QED
          BY <4>1, <4>2
      (*********************************************************************)
      (* The mapping from LL1ObservedOutputs to HLObservedOutputs is       *)
      (* direct.                                                           *)
      (*********************************************************************)
      <3>5. UNCHANGED HLObservedOutputs
        <4>1. UNCHANGED LL1ObservedOutputs
          BY <3>2
        <4>2. /\ HLObservedOutputs = LL1ObservedOutputs
              /\ HLObservedOutputs' = LL1ObservedOutputs'
          BY <2>1 DEF LL1Refinement
        <4>3. QED
          BY <4>1, <4>2
      (*********************************************************************)
      (* We prove the stuttering of the high-level public and private      *)
      (* state by using the NonAdvancementLemma.                           *)
      (*********************************************************************)
      <3>6. UNCHANGED <<HLPublicState, HLPrivateState>>
        (*******************************************************************)
        (* Many of the antecedents for the NonAdvancementLemma come        *)
        (* directly from antecedents in the induction.                     *)
        (*******************************************************************)
        <4>1. /\  LL1Refinement
              /\  LL1Refinement'
              /\  LL1TypeInvariant
              /\  LL1TypeInvariant'
              /\  UniquenessInvariant
          BY <2>1 DEF CorrectnessInvariants
        (*******************************************************************)
        (* The NVRAM is unchanged by definition of the LL1CorruptRAM       *)
        (* action.                                                         *)
        (*******************************************************************)
        <4>2. UNCHANGED LL1NVRAM
          BY <3>2
        (*******************************************************************)
        (* The UnchangedAuthenticatedHistoryStateBindingsLemma tells us     *)
        (* that there is no change to the set of history state bindings     *)
        (* that have authenticators in the set                            *)
        (* LL1ObservedAuthenticators.                                *)
        (*******************************************************************)
        <4>3. \A historyStateBinding \in HashType :
                  UNCHANGED LL1HistoryStateBindingAuthenticated(historyStateBinding)
          <5>1. UNCHANGED <<LL1NVRAM, LL1ObservedAuthenticators>>
            BY <3>2
          <5>2. QED
            BY <5>1, UnchangedAuthenticatedHistoryStateBindingsLemma
        (*******************************************************************)
        (* We have all of the antecedents for the NonAdvancementLemma, so  *)
        (* we can apply it directly.                                       *)
        (*******************************************************************)
        <4>4. QED
          BY <4>1, <4>2, <4>3, NonAdvancementLemma
      <3>7. QED
        BY <3>3, <3>4, <3>5, <3>6 DEF HLVars
    (***********************************************************************)
    (* A `^\basicname^' LL1RestrictedCorruption action refines to a        *)
    (* high-level HLDie step.                                              *)
    (***********************************************************************)
    <2>10. LL1RestrictedCorruption => HLDie
      <3>1. HAVE LL1RestrictedCorruption
      <3>2. PICK garbageHistorySummary \in HashType :
                 LL1RestrictedCorruption!nvram!(garbageHistorySummary)
        BY <3>1 DEF LL1RestrictedCorruption
      (*********************************************************************)
      (* First, we prove that this action causes the                       *)
      (* LL1NVRAMHistorySummaryUncorrupted predicate to become false.       *)
      (*********************************************************************)
      <3>3. LL1NVRAMHistorySummaryUncorrupted' = FALSE
        (*******************************************************************)
        (* We will make use of the conjunct in LL1RestrictedCorruption     *)
        (* that prevents the garbage history summary from being in an       *)
        (* authenticated history state binding.  This conjunct states a     *)
        (* 2-way universally quantifiied predicate.                        *)
        (*******************************************************************)
        <4>1. LL1RestrictedCorruption!nvram!current(garbageHistorySummary)
          BY <3>2
        (*******************************************************************)
        (* The following equivalence, plus the knowledge that the          *)
        (* symmetric key in the NVRAM and the set of observed        *)
        (* authenticators have not changed, are sufficient to prove that  *)
        (* the above 2-way universally quantified predicate equals the     *)
        (* negation of the 2-way universally quantified predicate in       *)
        (* LL1NVRAMHistorySummaryUncorrupted, when expanded through         *)
        (* LL1HistoryStateBindingAuthenticated.                             *)
        (*******************************************************************)
        <4>2. LL1NVRAM.historySummary' = garbageHistorySummary
          <5>1. LL1NVRAM' = [ historySummary |-> garbageHistorySummary,
                           symmetricKey |-> LL1NVRAM.symmetricKey ]
            BY <3>2
          <5>2. QED
            BY <5>1
        <4>3. UNCHANGED LL1NVRAM.symmetricKey
          BY <2>1, SymmetricKeyConstantLemma
        <4>4. UNCHANGED LL1ObservedAuthenticators
          BY <3>1 DEF LL1RestrictedCorruption
        <4>5. QED
          BY <4>1, <4>2, <4>3, <4>4
          DEF LL1NVRAMHistorySummaryUncorrupted, LL1HistoryStateBindingAuthenticated
      (*********************************************************************)
      (* Because the LL1NVRAMHistorySummaryUncorrupted is false, the        *)
      (* LL1Refinement immediately tells us that the high-level system is  *)
      (* not alive.                                                        *)
      (*********************************************************************)
      <3>4. HLAlive' = FALSE
        BY <2>1, <3>3 DEF LL1Refinement
      (*********************************************************************)
      (* The mapping from LL1AvailableInputs to HLAvailableInputs is       *)
      (* direct.                                                           *)
      (*********************************************************************)
      <3>5. UNCHANGED HLAvailableInputs
        <4>1. UNCHANGED LL1AvailableInputs
          BY <3>1 DEF LL1RestrictedCorruption
        <4>2. /\ HLAvailableInputs = LL1AvailableInputs
              /\ HLAvailableInputs' = LL1AvailableInputs'
          BY <2>1 DEF LL1Refinement
        <4>3. QED
          BY <4>1, <4>2
      (*********************************************************************)
      (* The mapping from LL1ObservedOutputs to HLObservedOutputs is       *)
      (* direct.                                                           *)
      (*********************************************************************)
      <3>6. UNCHANGED HLObservedOutputs
        <4>1. UNCHANGED LL1ObservedOutputs
          BY <3>1 DEF LL1RestrictedCorruption
        <4>2. /\ HLObservedOutputs = LL1ObservedOutputs
              /\ HLObservedOutputs' = LL1ObservedOutputs'
          BY <2>1 DEF LL1Refinement
        <4>3. QED
          BY <4>1, <4>2
      (*********************************************************************)
      (* Because the LL1NVRAMHistorySummaryUncorrupted is false, the        *)
      (* LL1Refinement immediately tells us that the high-level public     *)
      (* state equals the dead state.                                      *)
      (*********************************************************************)
      <3>7. HLPublicState' = DeadPublicState
        BY <2>1, <3>3 DEF LL1Refinement
      (*********************************************************************)
      (* Because the LL1NVRAMHistorySummaryUncorrupted is false, the        *)
      (* LL1Refinement immediately tells us that the high-level private    *)
      (* state equals the dead state.                                      *)
      (*********************************************************************)
      <3>8. HLPrivateState' = DeadPrivateState
        BY <2>1, <3>3 DEF LL1Refinement
      <3>9. QED
        BY <3>4, <3>5, <3>6, <3>7, <3>8 DEF HLDie
    <2>11. QED
      BY <2>1, <2>2, <2>3, <2>4, <2>5, <2>6, <2>7, <2>8, <2>9, <2>10 DEF HLNext, LL1Next
  <1>5. QED
    (***********************************************************************)
    (* Using the StepSimulation proof rule, the base case and the          *)
    (* induction step together imply that the invariant always holds.      *)
    (***********************************************************************)
    <2>1. [][LL1Next]_LL1Vars /\ []LL1Refinement /\ []CorrectnessInvariants => [][HLNext]_HLVars
      BY <1>4, StepSimulation
    <2>2. QED
      BY <1>1, <1>2, <1>3, <2>1 DEF LL1Spec, HLSpec, LL1Refinement

====
