---- MODULE MemoirLL1Refinement ----

(***************************************************************************)
(* This module describes how to interpret the state of the `^\basicname^'  *)
(* spec as a state of the high-level spec.                                 *)
(*                                                                         *)
(* This module includes the following definitions:                         *)
(*     LL1HistoryStateBindingAuthenticated                                 *)
(*     LL1NVRAMHistorySummaryUncorrupted                                   *)
(*     LL1Refinement                                                       *)
(***************************************************************************)

EXTENDS MemoirLL1Specification

(***************************************************************************)
(* The LL1HistoryStateBindingAuthenticated predicate asserts that a        *)
(* history state binding is authenticated, meaning that the set of         *)
(* observed authenticators includes an authenticator that is a valid MAC   *)
(* of this history state binding.                                          *)
(***************************************************************************)

LL1HistoryStateBindingAuthenticated(historyStateBinding) == \*`^\label{DEF-LL1HistoryStateBindingAuthenticated}^'
    \E authenticator \in LL1ObservedAuthenticators :
        ValidateMAC(LL1NVRAM.symmetricKey, historyStateBinding, authenticator)

(***************************************************************************)
(* The LL1NVRAMHistorySummaryUncorrupted predicate asserts that there      *)
(* exists some state hash that is bound to the history summary in the      *)
(* NVRAM by an authenticated history state binding.  This predicate is     *)
(* initially true, and it remains true until a LL1RestrictedCorruption     *)
(* action occurs, which makes the predicate false, and it remains false    *)
(* thereafter.                                                             *)
(***************************************************************************)

LL1NVRAMHistorySummaryUncorrupted == \*`^\label{DEF-LL1NVRAMHistorySummaryUncorrupted}^'
    \E stateHash \in HashType :
        LET
            historyStateBinding == Hash(LL1NVRAM.historySummary, stateHash)
        IN
            LL1HistoryStateBindingAuthenticated(historyStateBinding)

(***************************************************************************)
(* The LL1Refinement describes the relationship between the `^\basicname^' *)
(* spec and the high-level spec.                                           *)
(***************************************************************************)

LL1Refinement == \*`^\label{DEF-LL1Refinement}^'
        (*******************************************************************)
        (* The high-level available inputs correspond exactly to the       *)
        (* `^\basicname^' available inputs, since both are abstractions    *)
        (* that model the availability of particular input values to the   *)
        (* user.                                                           *)
        (*******************************************************************)
    /\  HLAvailableInputs = LL1AvailableInputs
        (*******************************************************************)
        (* The high-level observed outputs correspond exactly to the       *)
        (* `^\basicname^' observed outputs, since both are abstractions    *)
        (* that model the set of outputs that the user has so far observed *)
        (* from the operation of the service.                              *)
        (*******************************************************************)
    /\  HLObservedOutputs = LL1ObservedOutputs
        (*******************************************************************)
        (* The high-level public and private state is defined in terms of  *)
        (* the history summary in the NVRAM.  There are two possibilities. *)
        (*******************************************************************)
    /\  IF LL1NVRAMHistorySummaryUncorrupted
        THEN
            (***************************************************************)
            (* First, if the set of observed authenticators contains an    *)
            (* authenticator that binds any state hash to the history      *)
            (* summary currently in the NVRAM, then the public and private *)
            (* state is any state of the legal type whose hash is so       *)
            (* bound.  In this case, the service is alive.                 *)
            (***************************************************************)
            LET
                refPrivateStateEnc == SymmetricEncrypt(LL1NVRAM.symmetricKey, HLPrivateState)
                refStateHash == Hash(HLPublicState, refPrivateStateEnc)
                refHistoryStateBinding == Hash(LL1NVRAM.historySummary, refStateHash)
            IN
                /\  HLPublicState \in PublicStateType
                /\  HLPrivateState \in PrivateStateType
                /\  LL1HistoryStateBindingAuthenticated(refHistoryStateBinding)
                /\  HLAlive = TRUE
        ELSE
            (***************************************************************)
            (* Second, if the set of observed authenticators does not      *)
            (* contain an authenticator that binds any state hash to the   *)
            (* history summary currently in the NVRAM, then the values of  *)
            (* the public and private state are equal to their dead        *)
            (* states, and the service is not alive.                       *)
            (***************************************************************)
            /\  HLPublicState = DeadPublicState
            /\  HLPrivateState = DeadPrivateState
            /\  HLAlive = FALSE

====
